import torch
from torch import nn


class PGD(object):
    def __init__(self, model: nn.Module, emb_name, epsilon=1., alpha=0.3):
        self.model = model
        self.emb_name = emb_name
        self.epsilon = epsilon
        self.alpha = alpha
        self.emb_backup = {}
        self.grad_backup = {}

    def attack(self, is_first_attack=False):
        for name, param in self.model.named_parameters():
            if param.requires_grad and self.emb_name in name:
                if is_first_attack:
                    self.emb_backup[name] = param.data.clone()
                norm = torch.norm(param.grad)
                if norm != 0:
                    r_at = self.alpha * param.grad / norm
                    param.data.add_(r_at)
                    param.data = self.project(name, param.data, self.epsilon)

    def restore(self):
        for name, param in self.model.named_parameters():
            if param.requires_grad and self.emb_name in name:
                assert name in self.emb_backup
                param.data = self.emb_backup[name]
        self.emb_backup = {}

    def project(self, param_name: str, param_data, epsilon: float):
        r = param_data - self.emb_backup[param_name]
        if torch.norm(r) > epsilon:
            r = epsilon * r / torch.norm(r)
        return self.emb_backup[param_name] + r

    def backup_grad(self):
        for name, param in self.model.named_parameters():
            if param.requires_grad and param.grad is not None:
                self.grad_backup[name] = param.grad.clone()

    def restore_grad(self):
        for name, param in self.model.named_parameters():
            if param.requires_grad and param.grad is not None:
                param.grad = self.grad_backup[name]



import os
import re
from collections import defaultdict
from typing import Dict, List, Optional

import jieba
import numpy as np
from nltk import word_tokenize, PorterStemmer, WordNetLemmatizer
from nltk.corpus import stopwords

default_chinese_stop_words_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'stop_words',
                                               'stop_words.txt')


def get_stop_words(filepath=default_chinese_stop_words_path):
    with open(filepath, 'r', encoding='utf-8') as f:
        stop_words = [str(line).strip() for line in f.readlines()]
    return stop_words


def chinese_tokenize(text,
                     stop_words: Optional[List[str]] = None,
                     stop_words_path=default_chinese_stop_words_path) -> List[str]:
    cleaned_text = re.sub(u"[，。 :,.；|-“”——_/nbsp+&;@、《》～（）())#O！：【】]", "",
                          text).strip().lower()

    split_words = jieba.lcut(cleaned_text)
    if stop_words is None:
        stop_words = get_stop_words(stop_words_path)
    return [word for word in split_words if word not in stop_words]


def english_tokenize(text: str) -> List[str]:
    text = text.lower()
    text = re.sub(r'\d+', '', text)
    remove_chars = '[0-9’!"#$%&\'()*+,-./:;<=>?@，。?★、…【】《》？“”‘’！[\\]^_`{|}~]+'
    text = re.sub(remove_chars, ' ', text)

    text = text.strip()
    stop_words = stopwords.words("english")
    tokens = word_tokenize(text)
    tokens = [token for token in tokens if token not in stop_words]

    # 提取词干
    stemmer = PorterStemmer()
    tokens = [stemmer.stem(token) for token in tokens]

    # 同义词替换
    lemmatizer = WordNetLemmatizer()
    tokens = [lemmatizer.lemmatize(token) for token in tokens]

    return tokens


def generate_frequency_vocabulary(texts):
    vocab = defaultdict(int)
    for sentence in texts:
        for word in sentence:
            vocab[word] += 1
    return vocab


def add_unknown_words(word_vector_dict: Dict[str, np.ndarray], frequency_vocab: Dict[str, int], min_df=1, k=32):
    """
    For words that occur in at least min_df documents, create a separate word vector.
    0.25 is chosen so the unknown vectors have (approximately) same variance as pre-trained ones
    """
    for word in frequency_vocab:
        if word not in word_vector_dict and frequency_vocab[word] >= min_df:
            word_vector_dict[word] = np.random.uniform(-0.25, 0.25, k)


def padding_vec_and_idx(word_vectors: np.ndarray, word_idx: Dict[str, int]):
    word_vectors = np.concatenate(
        [np.zeros((1, word_vectors.shape[1]), dtype='float32'), word_vectors])
    word_idx_map = {word: index + 1 for word, index in word_idx.items()}
    return word_vectors, word_idx_map



import argparse


def fun(x=None, y=None):
    print(x, y)


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument()
    args = parser.parse_args()
    print(args.__dict__)
    fun(**args.__dict__)



import torch
import pandas as pd
import numpy as np
import transformers
import torchvision
from torchvision import transforms
from PIL import Image
from skimage import io, transform
from torch.utils.data import Dataset, DataLoader
from transformers import BertTokenizer

import torch.nn.functional as F
from transformers import BertModel
import random
import time
import os
import re

# 预处理
def text_preprocessing(text):
    """
    - 删除实体@符号(如。“@united”)
    — 纠正错误(如:'&amp;' '&')
    @参数 text (str):要处理的字符串
    @返回 text (Str):已处理的字符串
    """
    # 去除 '@name'
    text = re.sub(r'(@.*?)[\s]', ' ', text)

    #  替换'&amp;'成'&'
    text = re.sub(r'&amp;', '&', text)

    # 删除尾随空格
    text = re.sub(r'\s+', ' ', text).strip()

    return text


class FakeNewsDataset(Dataset):

    def __init__(self, df, root_dir, image_transform, tokenizer, MAX_LEN):
        """
        参数:
            csv_file (string):包含文本和图像名称的csv文件的路径
            root_dir (string):目录
            transform(可选):图像变换
        """
        self.csv_data = df
        self.root_dir = root_dir
        self.image_transform = image_transform
        self.tokenizer_bert = tokenizer
        self.MAX_LEN = MAX_LEN

    def __len__(self):
        return self.csv_data.shape[0]
    
    def pre_processing_BERT(self, sent):
        
        encoded_sent = self.tokenizer_bert.encode_plus(
            text=text_preprocessing(sent),  # 预处理
            add_special_tokens=True,        # `[CLS]`&`[SEP]`
            max_length=self.MAX_LEN,        # 截断/填充的最大长度
            padding='max_length',           # 句子填充最大长度
            # return_tensors='pt',          # 返回tensor
            return_attention_mask=True,     # 返回attention mask
            truncation=True
            )
        
        input_ids = encoded_sent.get('input_ids')
        attention_mask = encoded_sent.get('attention_mask')
        
        # 转换tensor
        input_ids = torch.tensor(input_ids)
        attention_mask = torch.tensor(attention_mask)
        
        return input_ids, attention_mask
     
        
    def __getitem__(self, idx):

        if torch.is_tensor(idx):
            idx = idx.tolist()
        
        img_name = self.root_dir + self.csv_data['image_id'][idx] + '.jpg'
        image = Image.open(img_name).convert("RGB")
        image = self.image_transform(image)
        
        text = self.csv_data['post_text'][idx]
        tensor_input_id, tensor_input_mask = self.pre_processing_BERT(text)

        label = self.csv_data['label'][idx]

        if label == 'fake':
            label = '1'
        else:
            label = '0'
        label = int(label)
        
        label = torch.tensor(label)

        sample = {
                  'image_id'  :  image, 
                  'BERT_ip'   : [tensor_input_id, tensor_input_mask],
                  'label'     :  label
                  }

        return sample


from typing import Optional

from torch import Tensor
from torch.utils.data import Dataset


class FinerFactDataset(Dataset):
    def __init__(self, token_id: Tensor, mask: Tensor, type_id: Tensor, label: Tensor, post_rank: Tensor,
                 user_rank: Tensor,
                 keyword_rank: Tensor, user_metadata: Optional[Tensor] = None):
        super().__init__()
        self.user_metadata = user_metadata
        self.keyword_rank = keyword_rank
        self.user_rank = user_rank
        self.post_rank = post_rank
        self.label = label
        self.type_id = type_id
        self.mask = mask
        self.token_id = token_id

    def __len__(self):
        return len(self.label)

    def __getitem__(self, index):
        item = {
            'token_id': self.token_id[index],
            'mask': self.mask[index],
            'type_id': self.type_id[index],
            'label': self.label[index],
            'post_rank': self.post_rank[index],
            'user_rank': self.user_rank[index],
            'keyword_rank': self.keyword_rank[index]
        }
        if self.user_metadata is not None:
            item['user_metadata'] = self.user_metadata[index]
        return item



from typing import Dict

import torch.nn.functional as F
from torch.utils.data import DataLoader

from faknow.train.trainer import BaseTrainer
from faknow.utils.pgd import PGD


class MFANTrainer(BaseTrainer):
    def _train_epoch(self, data: DataLoader, epoch: int) -> Dict[str, float]:
        self.model.train()

        pgd_word = PGD(self.model, emb_name='word_embedding', epsilon=6, alpha=1.8)
        losses = loss_defence = loss_adv = None

        for batch_id, batch_data in enumerate(data):
            # common loss
            losses = self.model.calculate_loss(batch_data)
            loss_defence = losses['total_loss']
            self.optimizer.zero_grad()
            loss_defence.backward()

            # PGD
            k = 3
            pgd_word.backup_grad()
            for t in range(k):
                pgd_word.attack(is_first_attack=(t == 0))
                if t != k - 1:
                    self.model.zero_grad()
                else:
                    pgd_word.restore_grad()
                y_pred = self.model.predict(batch_data)
                loss_adv = F.cross_entropy(y_pred, batch_data['label'])
                loss_adv.backward()
            pgd_word.restore()

            self.optimizer.step()

        return {k: v.item() for k, v in losses.items()}



from typing import Dict

import torch
import torch.nn as nn
import torch.nn.functional as F
import torchvision
from torch import Tensor

from faknow.model.layers.layer import GradientReverseLayer, TextCNNLayer
from faknow.model.model import AbstractModel

"""
EANN: Multi-Modal Fake News Detection
paper: https://dl.acm.org/doi/abs/10.1145/3219819.3219903
code: https://github.com/yaqingwang/EANN-KDD18
"""


class EANN(AbstractModel):
    r"""EANN: Multi-Modal Fake News Detection

        Args:
            event_num (int): number of events
            embed_weight (Tensor): weight for word embedding layer, shape=(vocab_size, embedding_size)
            reverse_lambda (float): lambda for gradient reverse layer. Default=1
            hidden_size (int): size for hidden layers. Default=32
        """

    def __init__(self,
                 event_num: int,
                 embed_weight: torch.Tensor,
                 reverse_lambda=1.0,
                 hidden_size=32):
        super(EANN, self).__init__()

        self.loss_funcs = [nn.CrossEntropyLoss(), nn.CrossEntropyLoss()]
        self.loss_weights = [1.0, 1.0]

        self.event_num = event_num
        self.embed_dim = embed_weight.shape[-1]
        self.hidden_size = hidden_size
        self.reverse_lambda = reverse_lambda

        # text
        self.embed = nn.Embedding.from_pretrained(embed_weight, freeze=False)

        filter_num = 20
        window_size = [1, 2, 3, 4]
        self.text_ccn_layer = TextCNNLayer(self.embed_dim, filter_num,
                                           window_size, F.leaky_relu)
        self.text_ccn_fc = nn.Linear(
            len(window_size) * filter_num, self.hidden_size)

        # image
        vgg_19 = torchvision.models.vgg19(
            weights=torchvision.models.VGG19_Weights.DEFAULT)
        for param in vgg_19.parameters():
            param.requires_grad = False

        self.vgg = vgg_19
        self.image_fc = nn.Linear(vgg_19.classifier._modules['6'].out_features, self.hidden_size)

        # Class Classifier
        self.class_classifier = nn.Linear(2 * self.hidden_size, 2)

        # Event Classifier
        self.domain_classifier = nn.Sequential(
            nn.Linear(2 * self.hidden_size, self.hidden_size),
            nn.LeakyReLU(True),
            nn.Linear(self.hidden_size, self.event_num))

    def forward(self, token_id: torch.Tensor,
                mask: torch.Tensor, image: torch.Tensor):
        """

        Args:
            token_id (Tensor): text token ids
            image (Tensor): image pixels
            mask (Tensor): text masks

        Returns:
            tuple:
                - class_output (Tensor): prediction of being fake news, shape=(batch_size, 2)
                - domain_output (Tensor): prediction of belonging to which domain, shape=(batch_size, domain_num)
        """
        # IMAGE
        image = self.vgg(image)  # [N, 512]
        image = F.leaky_relu(self.image_fc(image))

        # text CNN
        text = self.embed(token_id)
        text = text * mask.unsqueeze(2).expand_as(text)
        text = self.text_ccn_layer(text)
        text = F.leaky_relu(self.text_ccn_fc(text))

        # combine Text and Image
        text_image = torch.cat((text, image), 1)

        # Fake or real
        class_output = self.class_classifier(text_image)
        # Domain (which Event)
        reverse_feature = GradientReverseLayer.apply(text_image,
                                                     self.reverse_lambda)
        domain_output = self.domain_classifier(reverse_feature)

        return class_output, domain_output

    def calculate_loss(self, data) -> Dict[str, Tensor]:
        token_id = data['text']['token_id']
        mask = data['text']['mask']
        image = data['image']
        event_label = data['domain']
        label = data['label']
        class_output, domain_output = self.forward(token_id, mask, image)
        class_loss = self.loss_funcs[0](class_output,
                                        label) * self.loss_weights[0]
        domain_loss = self.loss_funcs[1](domain_output,
                                         event_label) * self.loss_weights[1]
        loss = class_loss + domain_loss

        return {'total_loss': loss, 'class_loss': class_loss, 'domain_loss': domain_loss}

    def predict(self, data_without_label) -> torch.Tensor:
        token_id = data_without_label['text']['token_id']
        mask = data_without_label['text']['mask']
        image = data_without_label['image']
        class_output, _ = self.forward(token_id, mask, image)
        return torch.softmax(class_output, dim=-1)



import torch
import torch.nn as nn
from torch import Tensor
from torchvision import models
from transformers import BertModel

from faknow.model.model import AbstractModel

"""
SpotFake: Multi-Modal Fake News Detection
paper: https://ieeexplore.ieee.org/document/8919302
code: https://github.com/shiivangii/SpotFake
"""


# 文本Bert基本模型
class _TextEncoder(nn.Module):
    def __init__(
            self,
            text_fc2_out=32,
            text_fc1_out=2742,
            dropout_p=0.4,
            fine_tune_module=False,
            pre_trained_bert_name="bert-base-uncased",
    ):
        super(_TextEncoder, self).__init__()
        self.fine_tune_module = fine_tune_module
        # 实例化
        self.bert = BertModel.from_pretrained(pre_trained_bert_name).requires_grad_(False)
        self.embedding_size = self.bert.config.hidden_size
        self.text_enc_fc1 = torch.nn.Linear(self.embedding_size, text_fc1_out)
        self.text_enc_fc2 = torch.nn.Linear(text_fc1_out, text_fc2_out)
        self.dropout = nn.Dropout(dropout_p)
        self.fine_tune()

    def forward(self, input_ids, attention_mask):
        """
        Args:
            input_ids (Tensor): text as input. shape=(batch_size, max_len)
            attention_mask (Tensor): attention mask information. shape=(batch_size, max_len)

        Returns:
            x (Tensor): encoded text. shape=(batch_size, text_fc2_out)
        """

        # 输入BERT
        out = self.bert(input_ids=input_ids, attention_mask=attention_mask)
        x = self.dropout(
            torch.nn.functional.relu(
                self.text_enc_fc1(out['pooler_output']))
        )
        x = self.dropout(
            torch.nn.functional.relu(
                self.text_enc_fc2(x))
        )
        return x

    def fine_tune(self):
        """
        固定参数
        """
        for p in self.bert.parameters():
            p.requires_grad = self.fine_tune_module


# 视觉vgg19预训练模型
class _VisionEncoder(nn.Module):
    def __init__(
            self,
            img_fc1_out=2742,
            img_fc2_out=32,
            dropout_p=0.4,
            fine_tune_module=False
    ):
        super(_VisionEncoder, self).__init__()
        self.fine_tune_module = fine_tune_module
        # 实例化
        vgg = models.vgg19(pretrained=True)
        vgg.classifier = nn.Sequential(*list(vgg.classifier.children())[:1])
        self.vis_encoder = vgg
        self.vis_enc_fc1 = torch.nn.Linear(4096, img_fc1_out)
        self.vis_enc_fc2 = torch.nn.Linear(img_fc1_out, img_fc2_out)
        self.dropout = nn.Dropout(dropout_p)
        self.fine_tune()

    def forward(self, images):
        """
        Args:
            images (Tensor): images as input. shape=(batch_size, 3, image_size, image_size)

        Returns:
            x (Tensor): encoded images. shape=(batch_size ,img_fc2_out)
        """

        x = self.vis_encoder(images)
        x = self.dropout(
            torch.nn.functional.relu(
                self.vis_enc_fc1(x))
        )
        x = self.dropout(
            torch.nn.functional.relu(
                self.vis_enc_fc2(x))
        )
        return x

    def fine_tune(self):
        """
        允许或阻止vgg的卷积块2到4的梯度计算。
        """
        for p in self.vis_encoder.parameters():
            p.requires_grad = False
        # 如果进行微调，则只微调卷积块2到4
        for c in list(self.vis_encoder.children())[5:]:
            for p in c.parameters():
                p.requires_grad = self.fine_tune_module


# LanguageAndVisionConcat
class _TextConcatVision(nn.Module):
    def __init__(
            self,
            model_params
    ):
        super(_TextConcatVision, self).__init__()
        self.text_encoder = _TextEncoder(model_params['text_fc2_out'], model_params['text_fc1_out'],
                                         model_params['dropout_p'], model_params['fine_tune_text_module'],
                                         model_params['pre_trained_bert_name'])
        self.vision_encoder = _VisionEncoder(model_params['img_fc1_out'], model_params['img_fc2_out'],
                                             model_params['dropout_p'], model_params['fine_tune_vis_module'])
        self.fusion = torch.nn.Linear(
            in_features=(model_params['text_fc2_out'] + model_params['img_fc2_out']),
            out_features=model_params['fusion_output_size']
        )
        self.fc = torch.nn.Linear(
            in_features=model_params['fusion_output_size'],
            out_features=1
        )
        self.dropout = torch.nn.Dropout(model_params['dropout_p'])

    def forward(self, text, image):
        """
        Args:
            text (Tensor): text as input. shape=[(batch_size, max_len), (batch_size, max_len)]
            image(Tensor): image as input. shape=(batch_size, 3, 224, 224)

        Returns:
            prediction (Tensor): prediction as output. shape=(8,)
        """
        # text to Bert
        text_features = self.text_encoder(text[0], text[1])
        # image to vgg
        image_features = self.vision_encoder(image)
        # 连接image & text
        combined_features = torch.cat(
            [text_features, image_features], dim=1
        )
        combined_features = self.dropout(combined_features)
        fused = self.dropout(
            torch.relu(
                self.fusion(combined_features)
            )
        )

        prediction = torch.sigmoid(self.fc(fused))
        prediction = prediction.squeeze(-1)
        prediction = prediction.float()
        return prediction


class SpotFake(AbstractModel):
    """
    SpotFake: A Multi-modal Framework for Fake News Detection
    """
    def __init__(
            self,
            text_fc2_out: int = 32,
            text_fc1_out: int = 2742,
            dropout_p: float = 0.4,
            fine_tune_text_module: bool = False,
            img_fc1_out: int = 2742,
            img_fc2_out: int = 32,
            fine_tune_vis_module: bool = False,
            fusion_output_size: int = 35,
            loss_func=nn.BCELoss(),
            pre_trained_bert_name="bert-base-uncased"
    ):
        """
        Args:
            text_fc2_out (int): size of the second fully connected layer of the text module. Default=32
            text_fc1_out (int): size of the first fully connected layer of the text module. Default=2742
            dropout_p (float): drop out rate. Default=0.4
            fine_tune_text_module (bool): text model fine-tuning or not. Default=False
            img_fc1_out (int): size of the first fully connected layer of the visual module. Default=2742
            img_fc2_out (int): size of the second fully connected layer of the visual module. Default=32
            fine_tune_vis_module (bool): visual model fine-tuning or not. Default=False
            fusion_output_size (int): size of the output layer after multimodal fusion. Default=35
            loss_func: loss function. Default=nn.BCELoss()
            pre_trained_bert_name: pretrained bert name. Default="bert-base-uncased"
        """

        super(SpotFake, self).__init__()
        self.text_fc2_out = text_fc2_out
        self.text_fc1_out = text_fc1_out
        self.dropout_p = dropout_p
        self.fine_tune_text_module = fine_tune_text_module
        self.img_fc1_out = img_fc1_out
        self.img_fc2_out = img_fc2_out
        self.fine_tune_vis_module = fine_tune_vis_module
        self.fusion_output_size = fusion_output_size
        self.pre_trained_bert_name = pre_trained_bert_name
        model_params = {
            "text_fc2_out": text_fc2_out,
            "text_fc1_out": text_fc1_out,
            "dropout_p": dropout_p,
            "fine_tune_text_module": fine_tune_text_module,
            "img_fc1_out": img_fc1_out,
            "img_fc2_out": img_fc2_out,
            "fine_tune_vis_module": fine_tune_vis_module,
            "fusion_output_size": fusion_output_size,
            "pre_trained_bert_name": pre_trained_bert_name
        }
        self.model = _TextConcatVision(model_params)
        if loss_func is None:
            self.loss_func = nn.BCELoss()
        else:
            self.loss_func = loss_func

    def forward(self, text: torch.Tensor, mask: torch.Tensor, domain: torch.Tensor):
        """
        Args:
            text (Tensor): shape=(batch_size, max_len)
            mask (Tensor): shape=(batch_size, max_len)
            domain (Tensor): shape=(batch_size, 3, 224, 224)

        Returns:
            self.model([text, mask], image=domain) (Tensor): shape=(8,)
        """
        return self.model([text, mask], image=domain)

    def calculate_loss(self, data) -> Tensor:
        img_ip, text_ip, label = data["image_id"], data["post_text"], data['label']
        b_input_ids = text_ip['input_ids']
        b_attn_mask = text_ip['attention_mask']
        imgs_ip = img_ip
        b_labels = label
        output = self.forward(b_input_ids, b_attn_mask, imgs_ip)
        return self.loss_func(output, b_labels.float())

    @torch.no_grad()
    def predict(self, data_without_label):
        img_ip, text_ip = data_without_label["image_id"], data_without_label["post_text"]
        b_input_ids = text_ip['input_ids']
        b_attn_mask = text_ip['attention_mask']
        imgs_ip = img_ip

        # shape=(n,), data = 1 or 0
        round_pred = self.forward(b_input_ids, b_attn_mask, imgs_ip)

        new_outputs = torch.zeros((round_pred.shape[0], 2)).to(round_pred.device)

        new_outputs[torch.where(round_pred < 0.5)[0].detach().cpu().numpy(), 0] = 1 - round_pred[round_pred < 0.5]
        new_outputs[torch.where(round_pred < 0.5)[0].detach().cpu().numpy(), 1] = round_pred[round_pred < 0.5]

        new_outputs[torch.where(round_pred >= 0.5)[0].detach().cpu().numpy(), 1] = round_pred[round_pred >= 0.5]
        new_outputs[torch.where(round_pred >= 0.5)[0].detach().cpu().numpy(), 0] = 1 - round_pred[round_pred >= 0.5]

        return new_outputs



import os
from typing import List, Callable, Any, Dict

import pandas as pd
import torch
from torch.utils.data import Dataset


class TextDataset(Dataset):
    def __init__(self,
                 path: str,
                 text_features: List[str],
                 tokenize: Callable[[List[str]], Any],
                 to_tensor=True):
        super().__init__()
        self.root = os.path.dirname(path)

        self.data = pd.read_json(path, orient='records').to_dict(orient='list')
        self.feature_names = list(self.data.keys())

        self.tokenize = tokenize
        self.text_features = {}
        for name in text_features:
            self.process_text(name)

        if to_tensor:
            self._to_tensor()

    def __getitem__(self, index) -> Dict[str, Any]:
        item = {}
        for feature_name, feature_values in self.data.items():
            if feature_name in self.text_features and type(
                    feature_values) is dict:
                value = {k: v[index] for k, v in feature_values.items()}
            else:
                value = feature_values[index]
            item[feature_name] = value
        return item

    def __len__(self):
        return len(self.data['label'])

    def check_feature(self, name: str):
        if name not in self.feature_names:
            raise ValueError(f"there is no feature named '{name}'")

    def process_text(self, name: str):
        self.check_feature(name)
        if name in self.text_features:
            raise ValueError(
                f"'{name}' has already been marked as text features")
        new_text = self.tokenize(self.data[name])
        if type(new_text) is dict:
            for k, v in new_text.items():
                if type(v) is not torch.Tensor:
                    raise TypeError(
                        f"the value of '{k}' returned by tokenize must be tensor"
                    )
            # todo 递归字典的情况，是否需要展开
            # self.data.update(new_text)
        elif type(new_text) is not torch.Tensor:
            raise TypeError("return type of tokenize function must be tensor")

        self.text_features[name] = self.data[name]  # 把旧值存储
        self.data[name] = new_text  # 更改为tokenize后的新值

    def remove_text(self, name: str):
        self.check_feature(name)
        if name not in self.text_features:
            raise ValueError(f"'{name}' has not been marked as text features")
        self.data[name] = self.text_features[name]  # 恢复旧值
        del self.text_features[name]  # 删除新值

    def _to_tensor(self):
        for name, values in self.data.items():
            if name not in self.text_features:
                if type(values[0]) is int or type(values[0]) is float:
                    try:
                        self.data[name] = torch.tensor(values)
                    except RuntimeError as e:
                        e.args = f"fail to convert '{name}' feature into tensor, please check data type"
                        raise e



from faknow.run.content_based.multimodal import *
from faknow.run.content_based.run_mdfend import run_mdfend, run_mdfend_from_yaml, TokenizerMDFEND
from faknow.run.content_based.run_textcnn import run_textcnn, run_textcnn_from_yaml, TokenizerTextCNN



import pickle
import re
from typing import Dict, List, Any

import jieba
import torch
import yaml
from PIL import Image
from torch.utils.data import DataLoader
from torchvision import transforms

from faknow.data.dataset.multi_modal import MultiModalDataset
from faknow.evaluate.evaluator import Evaluator
from faknow.model.content_based.multi_modal.eann import EANN
from faknow.train.trainer import BaseTrainer
from faknow.utils.util import dict2str, read_stop_words

__all__ = ['TokenizerEANN', 'transform_eann', 'adjust_lr_eann', 'run_eann', 'run_eann_from_yaml']


class TokenizerEANN:
    def __init__(self, vocab: Dict[str, int], max_len=255, stop_words: List[str] = None, language='zh') -> None:
        assert language in ['zh', 'en'], "language must be one of {zh, en}"
        self.language = language
        self.vocab = vocab
        self.max_len = max_len
        if stop_words is None:
            stop_words = []
        self.stop_words = stop_words

    def __call__(self, texts: List[str]) -> Dict[str, torch.Tensor]:
        token_ids = []
        masks = []
        for text in texts:
            cleaned_text = re.sub(u"[，。 :,.；|-“”——_/nbsp+&;@、《》～（）())#O！：【】]", "", text).strip().lower()

            if self.language == 'zh':
                split_words = jieba.cut_for_search(cleaned_text)
                words = " ".join([word for word in split_words if word not in self.stop_words])
            else:
                words = [word for word in cleaned_text.split() if word not in self.stop_words]

            token_id = [self.vocab[word] for word in words]
            real_len = len(token_id)
            if real_len < self.max_len:
                token_id.extend([0] * (self.max_len - real_len))
            elif real_len > self.max_len:
                token_id = token_id[:self.max_len]
            token_ids.append(token_id)

            mask = torch.zeros(self.max_len)
            mask[:real_len] = 1
            masks.append(mask)

        return {'token_id': torch.tensor(token_ids), 'mask': torch.stack(masks)}


def transform_eann(path: str) -> torch.Tensor:
    with open(path, "rb") as f:
        img = Image.open(f).convert('RGB')
        trans = transforms.Compose([
            transforms.Resize(256),
            transforms.CenterCrop(224),
            transforms.ToTensor(),
            transforms.Normalize([0.485, 0.456, 0.406], [0.229, 0.224, 0.225])
        ])
        return trans(img)


def adjust_lr_eann(epoch: int) -> float:
    return 0.001 / (1. + 10 * (float(epoch) / 100)) ** 0.75


def run_eann(train_path: str,
             vocab: Dict[str, int],
             stop_words: List[str],
             word_vectors: torch.Tensor,
             language='zh',
             max_len=255,
             batch_size=100,
             event_num: int = None,
             lr=0.001,
             num_epochs=100,
             metrics: List = None,
             validate_path: str = None,
             test_path: str = None,
             device='cpu') -> None:
    """
    todo run函数是否不要加过多参数，直接最基础最原始的几个参数就可以了，不需要很复杂，对于复杂功能就让用户自己写代码
    直接传入数据
    """

    tokenizer = TokenizerEANN(vocab, max_len, stop_words, language)

    # todo 是否在函数中加入validation_size，允许从train_path中划分出一部分作为validation set
    train_set = MultiModalDataset(train_path, ['text'], tokenizer, ['image'], transform_eann)
    train_loader = DataLoader(train_set, batch_size, shuffle=True)
    if event_num is None:
        event_num = torch.max(train_set.data['domain']).item() + 1

    if validate_path is not None:
        val_set = MultiModalDataset(validate_path, ['text'], tokenizer, ['image'], transform_eann)
        val_loader = DataLoader(val_set, batch_size, shuffle=False)
    else:
        val_loader = None

    model = EANN(event_num, embed_weight=word_vectors)
    optimizer = torch.optim.Adam(filter(lambda p: p.requires_grad,
                                        list(model.parameters())),
                                 lr)
    scheduler = torch.optim.lr_scheduler.LambdaLR(optimizer,
                                                  lr_lambda=adjust_lr_eann)
    evaluator = Evaluator(metrics)
    trainer = BaseTrainer(model, evaluator, optimizer, scheduler, device=device)
    trainer.fit(train_loader, num_epochs, validate_loader=val_loader)

    if test_path is not None:
        test_set = MultiModalDataset(test_path, ['text'], tokenizer, ['image'], transform_eann)
        test_loader = DataLoader(test_set, batch_size, shuffle=False)
        test_result = trainer.evaluate(test_loader)
        print(f"test result: {dict2str(test_result)}")


def _run_eann_from_config(config: Dict[str, Any]):
    run_eann(**_parse_kargs(config))


def _parse_kargs(config: Dict[str, Any]) -> Dict[str, Any]:
    with open(config['vocab'], 'rb') as f:
        config['vocab'] = pickle.load(f)
    with open(config['word_vectors'], 'rb') as f:
        config['word_vectors'] = pickle.load(f)
    config['stop_words'] = read_stop_words(config['stop_words'])
    return config


def run_eann_from_yaml(path: str):
    with open(path, 'r', encoding='utf-8') as _f:
        _config = yaml.load(_f, Loader=yaml.FullLoader)
        run_eann(**_parse_kargs(_config))



import math
from typing import Dict, Any

import torch
import yaml
from torch.optim import AdamW
from torch.utils.data import TensorDataset
from transformers import get_linear_schedule_with_warmup

from faknow.data.dataset.finerfact_dataset import FinerFactDataset
from faknow.evaluate.evaluator import Evaluator
from faknow.model.knowledge_aware.finerfact import FinerFact
from faknow.train.trainer import BaseTrainer
from faknow.utils.util import dict2str

__all__ = ['run_finerfact', 'run_finerfact_from_yaml']


def run_finerfact(train_data,
                  bert='bert-base-uncased',
                  test_data=None,
                  val_data=None,
                  lr=5e-5,
                  batch_size=8,
                  num_epochs=20,
                  gradient_accumulation_steps=8,
                  warmup_ratio=0.6,
                  metrics=None,
                  device='cpu'):
    train_set = FinerFactDataset(*train_data)

    train_loader = torch.utils.data.DataLoader(train_set,
                                               batch_size=batch_size,
                                               shuffle=True)
    if val_data is not None:
        val_set = FinerFactDataset(*val_data)
        val_loader = torch.utils.data.DataLoader(val_set,
                                                 batch_size=batch_size,
                                                 shuffle=False)
    else:
        val_loader = None

    model = FinerFact(bert)

    # optimizer
    named_params = list(model.named_parameters())
    no_decay = ['bias', 'LayerNorm.bias', 'LayerNorm.weight']
    optimizer_grouped_parameters = [{
        'params':
            [p for n, p in named_params if not any(nd in n for nd in no_decay)],
        'weight_decay':
            0.01
    }, {
        'params':
            [p for n, p in named_params if any(nd in n for nd in no_decay)],
        'weight_decay':
            0.0
    }]
    optimizer = AdamW(optimizer_grouped_parameters, lr=lr)

    # scheduler
    t_total = int(len(train_loader) / gradient_accumulation_steps * num_epochs)
    warmup_steps = math.ceil(t_total * warmup_ratio)
    scheduler = get_linear_schedule_with_warmup(optimizer,
                                                num_warmup_steps=warmup_steps,
                                                num_training_steps=t_total)

    evaluator = Evaluator(metrics)
    trainer = BaseTrainer(model, evaluator, optimizer, scheduler, device=device)
    trainer.fit(train_loader, num_epochs=num_epochs, validate_loader=val_loader)

    if test_data is not None:
        test_set = FinerFactDataset(*test_data)
        test_loader = torch.utils.data.DataLoader(test_set,
                                                  batch_size=batch_size,
                                                  shuffle=False)
        test_result = trainer.evaluate(test_loader)
        print(f"test result: {dict2str(test_result)}")


def _load_data(path: str):
    token_ids, masks, type_ids, labels, R_p, R_u, R_k, user_metadata, user_embeds = torch.load(
        path)
    return token_ids, masks, type_ids, labels, R_p, R_u, R_k, user_metadata, user_embeds


def _parse_kargs(config: Dict[str, Any]) -> Dict[str, Any]:
    config['train_data'] = _load_data('train_data')
    config['test_data'] = _load_data('test_data')
    return config


def run_finerfact_from_yaml(path: str):
    with open(path, 'r', encoding='utf-8') as _f:
        _config = yaml.load(_f, Loader=yaml.FullLoader)
        run_finerfact(**_parse_kargs(_config))



import torch
from sklearn.metrics import precision_score, recall_score, f1_score
from typing import Callable


def get_metric_func(name: str) -> Callable:
    name = name.lower()
    if name == 'accuracy':
        return calculate_accuracy
    if name == 'precision':
        return calculate_precision
    if name == 'recall':
        return calculate_recall
    if name == 'f1':
        return calculate_f1
    raise RuntimeError(f'no metric function called {name}')


def calculate_accuracy(outputs: torch.Tensor, y: torch.Tensor):
    return (outputs.argmax(dim=1).detach().cpu() == y.cpu()).float().mean().item()


def calculate_precision(outputs: torch.Tensor, y: torch.Tensor):
    return precision_score(outputs.argmax(dim=1).detach().cpu().numpy(), y.cpu().numpy(), zero_division=0)


def calculate_recall(outputs: torch.Tensor, y: torch.Tensor):
    return recall_score(outputs.argmax(dim=1).detach().cpu().numpy(), y.cpu().numpy(), zero_division=0)


def calculate_f1(outputs: torch.Tensor, y: torch.Tensor):
    return f1_score(outputs.argmax(dim=1).detach().cpu().numpy(), y.cpu().numpy(), zero_division=0)



from typing import Optional, List, Dict

import torch
import torch.nn.functional as F
from torch import Tensor, nn

from faknow.model.layers.layer import TextCNNLayer
from faknow.model.model import AbstractModel

"""
SAFE: Similarity-Aware Multi-Modal Fake News Detection
paper: https://dl.acm.org/doi/abs/10.1007/978-3-030-47436-2_27
code: https://github.com/Jindi0/SAFE
"""


def loss_func(cos_dis_sim: Tensor, label: Tensor) -> Tensor:
    return -(F.one_hot(label.long(), num_classes=2).float() * cos_dis_sim.log()).sum(1).mean()


class _TextCNN(nn.Module):
    def __init__(self,
                 input_size: int,
                 filter_num: int,
                 kernel_sizes: List[int],
                 dropout: float,
                 output_size: int):
        super().__init__()
        self.text_ccn_layer = TextCNNLayer(input_size, filter_num, kernel_sizes)
        self.dropout = nn.Dropout(dropout)
        self.fc = nn.Linear(len(kernel_sizes) * filter_num, output_size)

    def forward(self, x: torch.Tensor):
        x = self.text_ccn_layer(x)
        x = self.dropout(x)
        return self.fc(x)


class SAFE(AbstractModel):
    """
    SAFE: Similarity-Aware Multi-Modal Fake News Detection
    """

    def __init__(
            self,
            embedding_size: int = 300,
            conv_in_size: int = 32,
            filter_num: int = 128,
            cnn_out_size: int = 200,
            dropout: float = 0.,
            loss_weights: Optional[List[float]] = None,
    ):
        """

        Args:
            embedding_size (int): embedding size of text.
            conv_in_size (int): number of in channels in TextCNN. Default=32
            filter_num (int): number of filters in TextCNN. Default=128
            cnn_out_size (int): output size of FC layer in TextCNN. Default=200
            dropout (float): drop out rate. Default=0.0
            loss_weights (List[float]): list of loss weights. Default=[1.0, 1.0]
        """
        super(SAFE, self).__init__()

        self.loss_funcs = [loss_func, loss_func]
        if loss_weights is None:
            loss_weights = [1.0, 1.0]
        self.loss_weights = loss_weights

        self.embedding_size = embedding_size

        self.reduce = nn.Linear(embedding_size, conv_in_size)

        filter_sizes = [3, 4]
        self.head_block = _TextCNN(conv_in_size, filter_num, filter_sizes, dropout, cnn_out_size)
        self.body_block = _TextCNN(conv_in_size, filter_num, filter_sizes, dropout, cnn_out_size)
        self.image_block = _TextCNN(conv_in_size, filter_num, filter_sizes, dropout, cnn_out_size)

        self.predictor = nn.Linear(cnn_out_size * 3, 2)

        self.__init_weights__()

    def __init_weights__(self):
        nn.init.trunc_normal_(self.reduce.weight, std=0.1)
        nn.init.constant_(self.reduce.bias, 0.1)
        nn.init.trunc_normal_(self.predictor.weight, std=0.1)
        nn.init.constant_(self.predictor.bias, 0.1)

    def forward(
            self,
            head: torch.Tensor,
            body: torch.Tensor,
            image: torch.Tensor,
    ):
        """

        Args:
            head (Tensor): embedded title of post, shape=(batch_size, title_len, embedding_size)
            body (Tensor): embedded content of post, shape=(batch_size, content_len, embedding_size)
            image (Tensor): embedded sentence converted from post image, shape=(batch_size, sentence_len, embedding_size)

        Returns:
            tuple:
                - class_output (Tensor): prediction of being fake news, shape=(batch_size, 2)
                - cos_dis_sim (Tensor): prediction of belonging to which domain, shape=(batch_size, 2)
        """
        head = self.reduce(head)
        body = self.reduce(body)
        image = self.reduce(image)

        headline_vectors = self.head_block(head)
        body_vectors = self.body_block(body)
        image_vectors = self.image_block(image)

        combine_images = torch.cat([image_vectors, image_vectors], dim=1)
        combine_texts = torch.cat([headline_vectors, body_vectors], dim=1)

        combine_images_norm = combine_images.norm(p=2, dim=1)
        combine_texts_norm = combine_texts.norm(p=2, dim=1)

        image_text = (combine_images * combine_texts).sum(1)

        # cos similarity
        cosine_similarity = (
                                    1 + (image_text / (combine_images_norm * combine_texts_norm + 1e-8))
                            ) / 2
        cosine_distance = 1 - cosine_similarity
        cos_dis_sim = torch.stack([cosine_distance, cosine_similarity], 1)

        cat_vectors = torch.cat([headline_vectors, body_vectors, image_vectors], dim=1)
        class_output = self.predictor(cat_vectors)
        class_output = torch.softmax(class_output, dim=-1)

        return class_output, cos_dis_sim

    def calculate_loss(self, data) -> Dict[str, Tensor]:
        headline = data['head']
        body = data['body']
        image = data['image']
        label = data['label']
        class_output, cos_dis_sim = self.forward(headline, body, image)

        class_loss = self.loss_funcs[0](class_output, label) * self.loss_weights[0]
        cos_dis_sim_loss = self.loss_funcs[1](cos_dis_sim, label) * self.loss_weights[1]

        loss = class_loss + cos_dis_sim_loss
        return {'total_loss': loss, 'class_loss': class_loss, 'cos_dis_sim_loss': cos_dis_sim_loss}

    def predict(self, data_without_label) -> torch.Tensor:
        head = data_without_label['head']
        body = data_without_label['body']
        image = data_without_label['image']
        class_output, _ = self.forward(head, body, image)
        return class_output



from typing import List

import torch
import yaml
from torch_geometric.datasets import UPFD
from torch_geometric.loader import DenseDataLoader
from torch_geometric.transforms import ToUndirected, ToDense

from faknow.evaluate.evaluator import Evaluator
from faknow.model.social_context.gnncl import GNNCL
from faknow.train.dense_gnn_trainer import DenseGNNTrainer

__all__ = ['run_gnncl', 'run_gnncl_from_yaml']


def run_gnncl(root: str,
              name: str,
              feature: str,
              splits=None,
              batch_size=128,
              max_nodes=500,
              lr=0.1,
              epochs=70,
              metrics: List = None,
              device='cpu'):
    if splits is None:
        splits = ['train', 'val', 'test']

    train_dataset = UPFD(root,
                         name,
                         feature,
                         'train',
                         transform=ToDense(max_nodes),
                         pre_transform=ToUndirected())
    train_loader = DenseDataLoader(train_dataset,
                                   batch_size=batch_size,
                                   shuffle=True)

    if 'val' in splits:
        val_dataset = UPFD(root,
                           name,
                           feature,
                           'val',
                           transform=ToDense(max_nodes),
                           pre_transform=ToUndirected())
        val_loader = DenseDataLoader(val_dataset,
                                     batch_size=batch_size,
                                     shuffle=False)
    else:
        val_loader = None

    model = GNNCL(train_dataset.num_features, max_nodes)
    optimizer = torch.optim.Adam(model.parameters(), lr)
    evaluator = Evaluator(metrics)

    trainer = DenseGNNTrainer(model, evaluator, optimizer, device=device)
    trainer.fit(train_loader, epochs, val_loader)

    if 'test' in splits:
        test_dataset = UPFD(root,
                            name,
                            feature,
                            'test',
                            transform=ToDense(max_nodes),
                            pre_transform=ToUndirected())
        test_loader = DenseDataLoader(test_dataset,
                                      batch_size=batch_size,
                                      shuffle=False)
        test_result = trainer.evaluate(test_loader)
        print(f'test result={test_result}')


def run_gnncl_from_yaml(path: str):
    with open(path, 'r', encoding='utf-8') as _f:
        _config = yaml.load(_f, Loader=yaml.FullLoader)
        run_gnncl(**_config)



import json
import pickle
import re
from typing import List, Dict, Any

import jieba
import torch
import yaml
from PIL import Image
from torch import Tensor
from torch.utils.data import DataLoader
from torchvision import transforms

from faknow.data.dataset.multi_modal import MultiModalDataset
from faknow.evaluate.evaluator import Evaluator
from faknow.model.content_based.multi_modal.mfan import MFAN
from faknow.train.pgd_trainer import MFANTrainer
from faknow.utils.util import dict2str

__all__ = ['TokenizerMFAN', 'transform_mfan', 'load_adj_matrix_mfan', 'run_mfan', 'run_mfan_from_yaml']


class TokenizerMFAN:
    def __init__(self, vocab: Dict[str, int], max_len=50, stop_words: List[str] = None, language='zh') -> None:
        self.vocab = vocab
        self.max_len = max_len
        self.language = language
        if stop_words is None:
            stop_words = []
        self.stop_words = stop_words

    def __call__(self, texts: List[str]) -> Tensor:
        token_ids = []
        for text in texts:
            text = re.sub(r",", " , ", text)
            text = re.sub(r"!", " ! ", text)
            text = re.sub(r"\(", " \( ", text)
            text = re.sub(r"\)", " \) ", text)
            text = re.sub(r"\?", " \? ", text)
            text = re.sub(r"\s{2,}", " ", text).strip().lower()

            if self.language == 'zh':
                words = [word for word in jieba.cut(text) if word not in self.stop_words]
            else:
                words = [word for word in text.split() if word not in self.stop_words]

            token_id = [self.vocab[word] for word in words]
            real_len = len(token_id)
            if real_len < self.max_len:
                # padding zero in the front
                token_id = [0] * (self.max_len - real_len) + token_id
            elif real_len > self.max_len:
                token_id = token_id[:self.max_len]
            token_ids.append(token_id)

        return torch.tensor(token_ids)


def transform_mfan(path: str) -> torch.Tensor:
    with open(path, "rb") as f:
        img = Image.open(f).convert('RGB')
        trans = transforms.Compose([
            transforms.Resize(256),
            transforms.CenterCrop(224),
            transforms.ToTensor(),
            transforms.Normalize(mean=[0.485, 0.456, 0.406],
                                 std=[0.229, 0.224, 0.225])
        ])
        return trans(img)


def load_adj_matrix_mfan(path: str, node_num: int):
    with open(path, 'r') as f:
        adj_dict = json.load(f)

    adj_matrix = torch.zeros(size=(node_num, node_num))
    for ori_node, des_nodes in adj_dict.items():
        adj_matrix[int(ori_node), des_nodes] = 1
    return adj_matrix


def run_mfan(train_path: str,
             node_embedding: torch.Tensor,
             node_num: int,
             adj_matrix: torch.Tensor,
             vocab: Dict[str, int],
             word_vectors: torch.Tensor,
             max_len=50,
             batch_size=64,
             num_epochs=20,
             lr=2e-3,
             metrics: List = None,
             validate_path: str = None,
             test_path: str = None,
             device='cpu'
             ):
    tokenize = TokenizerMFAN(vocab, max_len)

    train_data = MultiModalDataset(train_path, ['text'], tokenize, ['image'], transform_mfan)
    train_loader = DataLoader(train_data, batch_size, True)

    if validate_path:
        validate_data = MultiModalDataset(validate_path, ['text'], tokenize, ['image'], transform_mfan)
        validate_loader = DataLoader(validate_data, batch_size, True)
    else:
        validate_loader = None

    model = MFAN(word_vectors, node_num, node_embedding, adj_matrix, device=device)
    optimizer = torch.optim.Adam(model.parameters(), lr)
    evaluator = Evaluator(metrics)
    trainer = MFANTrainer(model, evaluator, optimizer)
    trainer.fit(train_loader, num_epochs, validate_loader)

    if test_path is not None:
        test_data = MultiModalDataset(test_path, ['text'], tokenize, ['image'], transform_mfan)
        test_loader = DataLoader(test_data, batch_size, True)
        test_result = trainer.evaluate(test_loader)
        print('test result: ', dict2str(test_result))


def _parse_kargs(config: Dict[str, Any]) -> Dict[str, Any]:
    with open(config['vocab'], 'rb') as f:
        config['vocab'] = pickle.load(f)
    with open(config['word_vectors'], 'rb') as f:
        config['word_vectors'] = pickle.load(f)
    with open(config['node_embedding'], 'rb') as f:
        config['node_embedding'] = pickle.load(f)
    config['adj_matrix'] = load_adj_matrix_mfan(config['adj_matrix'], config['node_num'])

    return config


def run_mfan_from_yaml(path: str):
    with open(path, 'r', encoding='utf-8') as _f:
        _config = yaml.load(_f, Loader=yaml.FullLoader)
        run_mfan(**_parse_kargs(_config))



from math import ceil
from typing import Optional

import torch
import torch.nn.functional as F
from torch import Tensor
from torch_geometric.nn import DenseSAGEConv, dense_diff_pool

from faknow.model.model import AbstractModel

"""
Graph neural networks with continual learning for fake news detection from social media
paper: https://arxiv.org/abs/2007.03316
code: https://github.com/safe-graph/GNN-FakeNews
"""


class _DenseGNN(torch.nn.Module):
    def __init__(self,
                 in_channels: int,
                 hidden_channels: int,
                 out_channels: int,
                 normalize=False,
                 fc=True):
        super(_DenseGNN, self).__init__()
        self.conv1 = DenseSAGEConv(in_channels, hidden_channels, normalize)
        self.bn1 = torch.nn.BatchNorm1d(hidden_channels)
        self.conv2 = DenseSAGEConv(hidden_channels, hidden_channels, normalize)
        self.bn2 = torch.nn.BatchNorm1d(hidden_channels)
        self.conv3 = DenseSAGEConv(hidden_channels, out_channels, normalize)
        self.bn3 = torch.nn.BatchNorm1d(out_channels)

        if fc is True:
            self.fc = torch.nn.Linear(2 * hidden_channels + out_channels,
                                      out_channels)
        else:
            self.fc = None

    def bn(self, i: int, x: Tensor):
        batch_size, num_nodes, num_channels = x.size()
        x = x.view(-1, num_channels)
        x = getattr(self, 'bn{}'.format(i))(x)
        x = x.view(batch_size, num_nodes, num_channels)
        return x

    def forward(self, x: Tensor, adj: Tensor, mask: Optional[Tensor] = None):
        x0 = x
        x1 = self.bn(1, F.relu(self.conv1(x0, adj, mask)))
        x2 = self.bn(2, F.relu(self.conv2(x1, adj, mask)))
        x3 = self.bn(3, F.relu(self.conv3(x2, adj, mask)))
        x = torch.cat([x1, x2, x3], dim=-1)
        if self.fc is not None:
            x = F.relu(self.fc(x))
        return x


class GNNCL(AbstractModel):
    def __init__(self, feature_size: int, max_nodes: int):
        """
        Args:
            feature_size (int): dimension of input node feature
            max_nodes (int): maximum number of nodes in a graph
        """
        super(GNNCL, self).__init__()

        num_nodes = ceil(0.25 * max_nodes)
        self.gnn1_pool = _DenseGNN(feature_size, 64, num_nodes)
        self.gnn1_embed = _DenseGNN(feature_size, 64, 64, fc=False)

        num_nodes = ceil(0.25 * num_nodes)
        self.gnn2_pool = _DenseGNN(3 * 64, 64, num_nodes)
        self.gnn2_embed = _DenseGNN(3 * 64, 64, 64, fc=False)

        self.gnn3_embed = _DenseGNN(3 * 64, 64, 64, fc=False)

        self.fc1 = torch.nn.Linear(3 * 64, 64)
        self.fc2 = torch.nn.Linear(64, 2)

    def forward(self, x: Tensor, adj: Tensor, mask: Optional[Tensor] = None):
        """

        Args:
            x (Tensor): node feature, shape (batch_size, max_nodes, feature_size)
            adj (Tensor): adjacency matrix, shape (batch_size, max_nodes, max_nodes)
            mask (Optional[Tensor], optional): mask for adjacency matrix, shape (batch_size, max_nodes). Default=None.

        Returns:
            Tuple[Tensor, Tensor, Tensor]:
                output (Tensor): prediction of being fake, shape (batch_size, 2).
                l (Tensor): loss of first pooling layer.
                e (Tensor): entropy of first pooling layer.
        """
        s = self.gnn1_pool(x, adj, mask)
        x = self.gnn1_embed(x, adj, mask)

        x, adj, l1, e1 = dense_diff_pool(x, adj, s, mask)

        s = self.gnn2_pool(x, adj)
        x = self.gnn2_embed(x, adj)

        x, adj, l2, e2 = dense_diff_pool(x, adj, s)

        x = self.gnn3_embed(x, adj)

        x = x.mean(dim=1)
        x = F.relu(self.fc1(x))
        out = self.fc2(x)
        return out, l1 + l2, e1 + e2

    def calculate_loss(self, data) -> torch.Tensor:
        output, _, _ = self.forward(data.x, data.adj, data.mask)
        loss_func = torch.nn.CrossEntropyLoss()
        loss = loss_func(output, data.y.view(-1))
        return loss

    def predict(self, data_without_label) -> torch.Tensor:
        output, _, _ = self.forward(data_without_label.x,
                                    data_without_label.adj,
                                    data_without_label.mask)
        return F.softmax(output, dim=1)



import random
import re
from typing import Dict, List

import numpy as np
import yaml
from PIL import Image
import torch
from torch import nn
from torch.utils.data import DataLoader
from torchvision import transforms
from transformers import BertTokenizer

from faknow.data.dataset.multi_modal import MultiModalDataset
from faknow.evaluate.evaluator import Evaluator
from faknow.model.content_based.multi_modal.spotfake import SpotFake
from faknow.train.trainer import BaseTrainer
from faknow.utils.util import dict2str

__all__ = ['text_preprocessing', 'TokenizerSpotFake', 'transform_spotfake', 'run_spotfake', 'run_spotfake_from_yaml']


def text_preprocessing(text):
    """
    - 删除实体@符号(如。“@united”)
    — 纠正错误(如:'&amp;' '&')
    @参数 text (str):要处理的字符串
    @返回 text (Str):已处理的字符串
    """
    # 去除 '@name'
    text = re.sub(r'(@.*?)[\s]', ' ', text)

    #  替换'&amp;'成'&'
    text = re.sub(r'&amp;', '&', text)

    # 删除尾随空格
    text = re.sub(r'\s+', ' ', text).strip()

    return text


class TokenizerSpotFake:
    def __init__(self, max_len, pre_trained_bert_name):
        self.max_len = max_len
        self.pre_trained_bert_name = BertTokenizer.from_pretrained(pre_trained_bert_name, do_lower_case=True)

    def __call__(self, texts: List[str]) -> Dict[str, torch.Tensor]:
        # 定义列表存储文本处理后的结果
        input_ids_ls = []
        attention_mask_ls = []

        for text in texts:
            encoded_sent = self.pre_trained_bert_name.encode_plus(
                text=text_preprocessing(text),  # 预处理
                add_special_tokens=True,  # `[CLS]`&`[SEP]`
                max_length=self.max_len,  # 截断/填充的最大长度
                padding='max_length',  # 句子填充最大长度
                # return_tensors='pt',          # 返回tensor
                return_attention_mask=True,  # 返回attention mask
                truncation=True
            )
            input_ids = encoded_sent.get('input_ids')
            attention_mask = encoded_sent.get('attention_mask')

            # 转换tensor
            input_ids = torch.tensor(input_ids)
            attention_mask = torch.tensor(attention_mask)

            # 添加到列表中去
            input_ids_ls.append(input_ids)
            attention_mask_ls.append(attention_mask)

        return {'input_ids': torch.stack(input_ids_ls), 'attention_mask': torch.stack(attention_mask_ls)}


def transform_spotfake(path: str) -> torch.Tensor:
    with open(path, "rb") as f:
        img = Image.open(f).convert('RGB')
        trans = transforms.Compose([
            transforms.Resize(size=(224, 224)),
            transforms.ToTensor(),
            transforms.Normalize([0.485, 0.456, 0.406], [0.229, 0.224, 0.225])
        ])
        return trans(img)


def run_spotfake(
        train_path: str,
        validate_path: str = None,
        test_path: str = None,
        text_fc2_out: int = 32,
        text_fc1_out: int = 2742,
        dropout_p: float = 0.4,
        fine_tune_text_module: bool = False,
        img_fc1_out: int = 2742,
        img_fc2_out: int = 32,
        fine_tune_vis_module: bool = False,
        fusion_output_size: int = 35,
        loss_func=nn.BCELoss(),
        pre_trained_bert_name="bert-base-uncased",
        batch_size=8,
        epochs=50,
        max_len=500,
        lr=3e-5,
        metrics: List = None,
        device='cuda:0'
):
    seed_value = 42
    random.seed(seed_value)
    np.random.seed(seed_value)
    torch.manual_seed(seed_value)
    torch.cuda.manual_seed_all(seed_value)

    tokenizer = TokenizerSpotFake(max_len, pre_trained_bert_name)

    training_set = MultiModalDataset(train_path, ['post_text'], tokenizer, ['image_id'], transform_spotfake)
    train_loader = DataLoader(training_set, batch_size=batch_size, shuffle=True)

    if validate_path is not None:
        validation_set = MultiModalDataset(validate_path, ['post_text'], tokenizer, ['image_id'], transform_spotfake)
        validation_loader = DataLoader(validation_set, batch_size=batch_size, shuffle=True)
    else:
        validation_loader = None

    model = SpotFake(text_fc2_out, text_fc1_out, dropout_p, fine_tune_text_module,
                     img_fc1_out, img_fc2_out, fine_tune_vis_module, fusion_output_size,
                     loss_func, pre_trained_bert_name)

    optimizer = torch.optim.AdamW(
        model.parameters(),
        lr
    )

    evaluator = Evaluator(metrics)

    trainer = BaseTrainer(model=model, evaluator=evaluator, optimizer=optimizer, device=device)
    trainer.fit(train_loader, epochs, validation_loader)

    if test_path is not None:
        test_set = MultiModalDataset(test_path, ['post_text'], tokenizer, ['image_id'], transform_spotfake)
        test_loader = DataLoader(test_set, batch_size, shuffle=False)
        test_result = trainer.evaluate(test_loader)
        print(f"test result: {dict2str(test_result)}")


def run_spotfake_from_yaml(path: str):
    with open(path, 'r', encoding='utf-8') as _f:
        _config = yaml.load(_f, Loader=yaml.FullLoader)
        run_spotfake(**_config)



from typing import Tuple, Any, Optional, Callable, Dict, List, Set

import torch.utils.data
import torchvision
from torchvision.datasets.folder import default_loader

from faknow.data.legacy.text_dataset import FolderTextDataset, TensorTextDataset
from faknow.data.dataset.utils import walker_with_images


class MultiModalDataset(torchvision.datasets.ImageFolder):

    # todo 是否需要label_transform
    def __init__(self,
                 img_root: str,
                 texts,
                 img_transform: Optional[Callable] = None,
                 loader: Callable[[str], Any] = default_loader,
                 other_params: Optional[Dict[str, Any]] = None):
        super().__init__(img_root, img_transform, loader=loader)
        self.texts = texts
        self.other_params = other_params

    def __getitem__(self, index: int) -> Tuple:
        img, label = super().__getitem__(index)
        text = self.texts[index]
        if self.other_params is None:
            return text, img, label
        other_params = {k: v[index] for k, v in self.other_params.items()}
        return text, img, other_params, label

    @property
    def data(self):
        return self.texts, [
            self.transform(img) if self.transform is not None else img
            for img in self.imgs
        ]


class TensorMultiModalDataset(TensorTextDataset):
    """
    inherit from TensorTextDataset for multi-modal data
    """

    def __init__(self,
                 texts: torch.Tensor = None,
                 images: torch.Tensor = None,
                 labels: torch.Tensor = None,
                 samples: Optional[List[Tuple[torch.Tensor,
                                              torch.Tensor,
                                              torch.Tensor]]] = None,
                 **other_data: torch.Tensor):
        """
        Args:
            texts:
            images:
            labels:
            samples: instead of inputting texts, image and labels, they can be combined as a list of tuples (text, image, label).
            other_data: other data the model needs, it should be torch tensor
        """
        super().__init__(texts, labels, samples, **other_data)

        if samples is not None:
            images = [sample[1] for sample in samples]
        self.images = images

        if len(self.texts) != len(self.images) or len(self.images) != len(
                self.labels) or len(self.texts) != len(self.labels):
            raise RuntimeError(
                'the first dimension among texts, images and labels must be the same'
            )

    def __getitem__(self, index: int) -> Tuple:
        item = super().__getitem__(index)
        if len(item) == 2:
            # text, image, label
            return item[0], self.images[index], item[1]
        # text, image, other_data, label
        return item[0], self.images[index], item[1], item[2]


class FolderMultiModalDataset(FolderTextDataset):
    def __init__(
            self,
            root: str,
            embedding: Callable[[str, Optional[Dict]], Any],
            transform: Optional[Callable],
            image_loader: Callable[[str], Any] = default_loader,
            is_valid_file: Optional[Callable[[str], bool]] = None,
            walk_class_dir: Optional[Callable[[str, str, int, List[Tuple], Set],
                                              None]] = walker_with_images,
            embedding_params: Dict[str, Any] = None,
            **other_data: torch.Tensor):
        super().__init__(root, embedding, walk_class_dir, embedding_params, **other_data)
        self.images = [sample[1] for sample in self.samples]
        self.transform = transform
        self.image_loader = image_loader

    def __getitem__(self, index: int) -> Tuple:
        item = super().__getitem__(index)
        image = self.image_loader(self.images[index])
        image = self.transform(image)

        if len(item) == 2:
            # text, image, label
            return item[0], image, item[1]
        # text, image, other_data, label
        return item[0], image, item[1], item[2]



from typing import Tuple

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch import Tensor
from transformers import BertModel

from faknow.model.model import AbstractModel

"""
Zoom Out and Observe: News Environment Perception for Fake News Detection
paper: https://aclanthology.org/2022.acl-long.311/
code: https://github.com/ICTMCG/News-Environment-Perception
"""


class _NewsEnvExtractor(nn.Module):
    def __init__(self,
                 bert_hidden_dim: int,
                 macro_env_out_dim=128,
                 micro_env_output_dim=128):
        """
        Args:
            bert_hidden_dim (int): bert hidden dimension
            macro_env_out_dim (int): macro environment output dimension, default=128
            micro_env_output_dim (int): micro environment output dimension, default=128
        """
        super(_NewsEnvExtractor, self).__init__()

        # gaussian kernel
        kernel_mu = torch.arange(-1, 1.1, 0.1).tolist()
        kernel_sigma = [20 for _ in kernel_mu]
        kernel_mu.append(0.99)
        kernel_sigma.append(100)
        self.kernel_mu = torch.tensor(kernel_mu)
        self.kernel_sigma = torch.tensor(kernel_sigma)

        # macro env
        self.macro_env_out_dim = macro_env_out_dim
        self.micro_env_output_dim = micro_env_output_dim

        macro_mlp_in = 0
        macro_mlp_in += 2 * bert_hidden_dim  # post与avg mac bert表示
        macro_mlp_in += len(self.kernel_mu)  # gaussian kernel

        self.macro_mlp = nn.Linear(macro_mlp_in, self.macro_env_out_dim)

        # micro env
        micro_output = 0
        # mic sem
        self.micro_sem_mlp = nn.Linear(2 * bert_hidden_dim,
                                       self.micro_env_output_dim)
        micro_output += self.micro_env_output_dim

        # mic sim
        self.micro_sim_mlp = nn.Linear(2 * len(self.kernel_mu),
                                       self.micro_env_output_dim)
        micro_output += self.micro_env_output_dim

        self.micro_mlp = nn.Linear(micro_output, self.micro_env_output_dim)

    def forward(self, post_simcse, avg_emac, avg_emic, kernel_p_emac,
                kernel_p_emic,
                kernel_avgmic_emic) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Args:
            post_simcse: post simcse representation, shape=(batch_size, simcse_dim)
            avg_emac: average macro news simcse representation, shape=(batch_size, simcse_dim)
            avg_emic: average micro news simcse representation, shape=(batch_size, simcse_dim)
            kernel_p_emac: gaussian kernel of similarity between post and macro news, shape=(batch_size, kernel_num)
            kernel_p_emic: gaussian kernel of similarity between  post and micro news, shape=(batch_size, kernel_num)
            kernel_avgmic_emic: gaussian kernel of similarity between average micro and micro news, shape=(batch_size, kernel_num)

        Returns:
            tuple:
                - macro_env: macro environment representation, shape=(batch_size, macro_env_out_dim)
                - micro_env: micro environment representation, shape=(batch_size, micro_env_out_dim)
        """
        p = post_simcse

        # ------------------ Macro Env ------------------
        # 宏观环境与post的高斯核，已事先计算好，只需norm
        kernel_p_emac = self.normalize(kernel_p_emac)

        # 三者concat作为最终宏观表征
        vectors = [x for x in [p, avg_emac, kernel_p_emac] if x is not None]
        v_p_mac = torch.cat(vectors, dim=-1)
        v_p_mac = self.macro_mlp(v_p_mac)

        # ------------------ Micro Env ------------------

        # mic sem
        # 平均值与post合并作为语义信息的聚合
        u_sem = torch.cat([p, avg_emic], dim=1)
        u_sem = self.micro_sem_mlp(u_sem)

        kernel_p_emic = self.normalize((kernel_p_emic))
        kernel_avgmic_emic = self.normalize(kernel_avgmic_emic)

        # mic sim
        u_sim = torch.cat([
            kernel_p_emic * kernel_avgmic_emic,
            kernel_p_emic - kernel_avgmic_emic
        ],
                          dim=1)
        u_sim = self.micro_sim_mlp(u_sim)

        vectors = [x for x in [u_sim, u_sem] if x is not None]
        v_p_mic = torch.cat(vectors, dim=-1)
        v_p_mic = self.micro_mlp(v_p_mic)

        return v_p_mac, v_p_mic

    def normalize(self, kernel_features):
        # Normalize
        zero = 1e-8
        kernel_sum = torch.sum(kernel_features)
        kernel_features /= (kernel_sum + zero)

        return kernel_features


class _BERT(nn.Module):
    def __init__(self,
                 bert: str,
                 max_len=256,
                 finetune_embedding_layers=True,
                 finetune_inter_layers=True):
        super(_BERT, self).__init__()

        self.bert = BertModel.from_pretrained(bert, return_dict=False)
        self.finetune(finetune_embedding_layers, finetune_inter_layers)

        self.max_len = max_len
        self.doc_max_len = self.max_len - 2
        self.out_dim = self.bert.config.hidden_size

    def finetune(self, finetune_embedding_layers: bool,
                 finetune_inter_layers: bool):
        for name, param in self.bert.named_parameters():
            # pooler layer
            if name.startswith("pooler"):
                if 'bias' in name:
                    param.data.zero_()
                elif 'weight' in name:
                    param.data.normal_(mean=0.0,
                                       std=self.bert.config.initializer_range)
                param.requires_grad = True

            # last encoder layer
            elif name.startswith('encoder.layer.11'):
                param.requires_grad = True

            # embedding layer
            elif name.startswith('embeddings'):
                param.requires_grad = finetune_embedding_layers

            # the other transformer layers (intermediate layers)
            else:
                param.requires_grad = finetune_inter_layers

    def forward(self, token: torch.Tensor):
        # (batch_size, max_length)
        input_ids, masks = self._encode(token)

        # (batch_size, max_length, bert_dim)
        seq_output, _ = self.bert(input_ids)

        # (batch_size, bert_dim)
        output = torch.sum(masks * seq_output, dim=1)

        return output

    def _encode(self, tokens: torch.Tensor):
        input_ids = []
        masks = []
        for token in tokens:
            doc = token[:self.doc_max_len].tolist()

            # add CLS[101], SEP[102], PAD[103]
            # [101, ..., 102, 103, 103, ..., 103]
            padding_length = self.max_len - (len(doc) + 2)
            input_id = [101] + doc + [102] + [103] * padding_length
            input_ids.append(input_id)

            mask = torch.zeros(self.max_len, 1, dtype=torch.float)
            mask[:-padding_length] = 1 / (len(doc) + 2)
            masks.append(mask)

        return torch.tensor(input_ids, dtype=torch.long), torch.stack(masks)


class NEP(AbstractModel):
    """
    Zoom Out and Observe: News Environment Perception for Fake News Detection
    """
    def __init__(self,
                 macro_env_out_dim=128,
                 micro_env_out_dim=128,
                 num_mlp_layers=3,
                 fnd: nn.Module = _BERT,
                 fusion='gate',
                 attention_dim=128,
                 **kwargs):
        """
        Args:
            macro_env_out_dim (int): dimension of macro environment representation, default=128
            micro_env_out_dim (int): dimension of micro environment representation, default=128
            num_mlp_layers (int): number of mlp layers, default=3
            fnd (nn.Module): fake news detector, default=_BERT
            fusion (str): fusion method to combine environment and fend, 'gate', 'att' or 'concat', default='gate'
            attention_dim (int): dimension of fusion attention layers, default=128
            **kwargs: parameters for fnd initialization
        """

        super(NEP, self).__init__()
        self.macro_env_out_dim = macro_env_out_dim
        self.micro_env_out_dim = micro_env_out_dim

        # === Env ===
        self.news_env_extractor = _NewsEnvExtractor(768, macro_env_out_dim,
                                                    micro_env_out_dim)

        # === FEND ===
        if fnd is not None:
            self.fake_news_detector = fnd(**kwargs)
            last_output = self.fake_news_detector.out_dim

            # Env and FEND fusion
            self.fusion = fusion

            if self.fusion == 'att':
                self.macro_multihead_attn = nn.MultiheadAttention(
                    attention_dim, num_heads=8, dropout=0.5)
                self.micro_multihead_attn = nn.MultiheadAttention(
                    attention_dim, num_heads=8, dropout=0.5)
                last_output += 2 * attention_dim
            elif self.fusion == 'gate':
                assert self.macro_env_out_dim == self.micro_env_out_dim
                self.W_gate = nn.Linear(
                    self.fake_news_detector.out_dim + self.macro_env_out_dim,
                    self.macro_env_out_dim)
                last_output += self.macro_env_out_dim
            elif self.fusion == 'concat':
                last_output += self.macro_env_out_dim + self.micro_env_output_dim
            else:
                raise ValueError('Invalid fusion strategy.')
        else:
            self.fake_news_detector = None
            last_output = 0

        # === MLP layers ===
        self.fcs = []
        for _ in range(num_mlp_layers - 1):
            curr_output = int(last_output / 2)
            self.fcs.append(nn.Linear(last_output, curr_output))
            last_output = curr_output
        self.fcs.append(nn.Linear(last_output, 2))
        self.fcs = nn.ModuleList(self.fcs)

    def forward(self, post_simcse, avg_mac, avg_mic, kernel_p_mac,
                kernel_p_mic, kernel_avg_mic_mic, **kwargs):
        """
        Args:
            post_simcse: post simcse representation, shape=(batch_size, simcse_dim)
            avg_mac: average macro news simcse representation, shape=(batch_size, simcse_dim)
            avg_mic: average micro news simcse representation, shape=(batch_size, simcse_dim)
            kernel_p_mac: gaussian kernel of similarity between post and macro news, shape=(batch_size, kernel_num)
            kernel_p_mic: gaussian kernel of similarity between  post and micro news, shape=(batch_size, kernel_num)
            kernel_avg_mic_mic: gaussian kernel of similarity between average micro and micro news, shape=(batch_size, kernel_num)
            **kwargs: parameters for fnd forward

        Returns:
            output: prediction of being fake, shape=(batch_size, 2)
        """
        # env
        v_p_mac, v_p_mic = self.news_env_extractor(post_simcse, avg_mac,
                                                   avg_mic, kernel_p_mac,
                                                   kernel_p_mic,
                                                   kernel_avg_mic_mic)

        if self.fake_news_detector is not None:
            # (bs, FEND_last_output)
            detector_output = self.fake_news_detector(**kwargs)

            # fusion
            output = None
            if self.fusion == 'concat':
                output = torch.cat([detector_output, v_p_mac, v_p_mic], dim=-1)
            elif self.fusion == 'att':
                output = self.forward_attention(detector_output, v_p_mac,
                                                v_p_mic)
            elif self.fusion == 'gate':
                output = self.forward_gate(detector_output, v_p_mac, v_p_mic)

        else:
            output = torch.cat([v_p_mac, v_p_mic], dim=-1)

        for fc in self.fcs:
            output = F.gelu(fc(output))

        return output

    def forward_attention(self,
                          detector_output=None,
                          v_p_mac=None,
                          v_p_mic=None):
        # (1, bs, 768)
        key = detector_output.unsqueeze(0)
        value = key

        # env_output: (1, bs, emb_dim), env_weights: (bs, 1, 1)
        macro_output, _ = self.macro_multihead_attn(query=v_p_mac.unsqueeze(0),
                                                    key=key,
                                                    value=value)
        micro_output, _ = self.micro_multihead_attn(query=v_p_mic.unsqueeze(0),
                                                    key=key,
                                                    value=value)

        output = torch.cat([
            detector_output,
            macro_output.squeeze(0),
            micro_output.squeeze(0)
        ],
                           dim=-1)
        return output

    def forward_gate(self, detector_output=None, v_p_mac=None, v_p_mic=None):
        # (bs, env_dim)
        g = torch.sigmoid(
            self.W_gate(torch.cat([detector_output, v_p_mac], dim=-1)))
        # (bs, env_dim)
        v_p = g * v_p_mac + (1 - g) * v_p_mic
        # (bs, env_dim + FEND_last_output)
        output = torch.cat([detector_output, v_p], dim=-1)
        return output

    def calculate_loss(self, data) -> Tensor:
        post_simcse, avg_mac, avg_mic, p_mac, p_mic, avg_mic_mic, token, label = data.values(
        )
        output = self(post_simcse,
                      avg_mac,
                      avg_mic,
                      p_mac,
                      p_mic,
                      avg_mic_mic,
                      token=token)
        criterion = nn.CrossEntropyLoss()
        loss = criterion(output, label)
        return loss

    def predict(self, data_without_label):
        post_simcse, avg_mac, avg_mic, p_mac, p_mic, avg_mic_mic, token, _ = data_without_label.values(
        )
        output = self(post_simcse,
                      avg_mac,
                      avg_mic,
                      p_mac,
                      p_mic,
                      avg_mic_mic,
                      token=token)
        return F.softmax(output, dim=-1)



from typing import Optional, List

import torch
import torch.nn.functional as F
from torch import nn, Tensor
from torchvision.models import vgg19, VGG19_Weights
from transformers import BertModel

from faknow.model.layers.dct import DctStem, DctInceptionBlock, conv2d_bn_relu
from faknow.model.layers.transformer import FFN, AddNorm
from faknow.model.model import AbstractModel

"""
Multimodal Fusion with Co-Attention Networks for Fake News Detection
paper: https://aclanthology.org/2021.findings-acl.226/
code: https://github.com/wuyang45/MCAN_code
"""


class _VGG(nn.Module):
    def __init__(self):
        super(_VGG, self).__init__()
        vgg_19 = vgg19(weights=VGG19_Weights.DEFAULT)

        self.feature = vgg_19.features
        self.classifier = nn.Sequential(*list(vgg_19.classifier.children())[:-3])

    def forward(self, img):
        img = self.feature(img)
        img = img.view(img.size(0), -1)
        image = self.classifier(img)

        return image


class _DctCNN(nn.Module):
    def __init__(self,
                 dropout,
                 kernel_sizes,
                 num_channels,
                 in_channel=128,
                 branch1_channels=None,
                 branch2_channels=None,
                 branch3_channels=None,
                 branch4_channels=None,
                 out_channels=64):
        super(_DctCNN, self).__init__()

        if branch4_channels is None:
            branch4_channels = [32]
        if branch3_channels is None:
            branch3_channels = [64, 96, 96]
        if branch2_channels is None:
            branch2_channels = [48, 64]
        if branch1_channels is None:
            branch1_channels = [64]

        self.stem = DctStem(kernel_sizes, num_channels)

        self.InceptionBlock = DctInceptionBlock(
            in_channel,
            branch1_channels,
            branch2_channels,
            branch3_channels,
            branch4_channels,
        )

        self.maxPool = nn.MaxPool2d((1, 122))

        self.dropout = nn.Dropout(dropout)

        self.conv = conv2d_bn_relu(branch1_channels[-1] + branch2_channels[-1] +
                                   branch3_channels[-1] + branch4_channels[-1],
                                   out_channels,
                                   kernel_size=1)

    def forward(self, dct_img):
        dct_f = self.stem(dct_img)
        x = self.InceptionBlock(dct_f)
        x = self.maxPool(x)
        x = x.permute(0, 2, 1, 3)
        x = self.conv(x)
        x = x.permute(0, 2, 1, 3)
        x = x.squeeze(-1)

        x = x.reshape(-1, 4096)

        return x


class _ScaledDotProductAttention(nn.Module):

    def __init__(self, attention_dropout=0.5):
        super(_ScaledDotProductAttention, self).__init__()
        self.dropout = nn.Dropout(attention_dropout)
        self.softmax = nn.Softmax(dim=2)

    def forward(self, q, k, v, scale=None):
        attention = torch.matmul(q, k.transpose(-2, -1))
        if scale:
            attention = attention * scale

        attention = self.softmax(attention)
        attention = self.dropout(attention)
        attention = torch.matmul(attention, v)

        return attention


class _MultiHeadAttention(nn.Module):
    """
    multi-head attention + add&norm
    """

    def __init__(self, model_dim=256, num_heads=8, dropout=0.5):
        super(_MultiHeadAttention, self).__init__()

        self.model_dim = model_dim
        self.dim_per_head = model_dim // num_heads
        self.num_heads = num_heads
        self.linear_k = nn.Linear(1, self.dim_per_head * num_heads, bias=False)
        self.linear_v = nn.Linear(1, self.dim_per_head * num_heads, bias=False)
        self.linear_q = nn.Linear(1, self.dim_per_head * num_heads, bias=False)

        self.dot_product_attention = _ScaledDotProductAttention(dropout)
        self.linear_final = nn.Linear(model_dim, 1, bias=False)
        self.dropout = nn.Dropout(dropout)
        self.layer_norm = nn.LayerNorm(model_dim)

    def forward(self, query, key, value):
        """
        Args:
            query: (batch_size, model_dim, 1)
            key: (batch_size, model_dim, 1)
            value: (batch_size, model_dim, 1)
            attn_mask: None

        Returns:
            shape=(batch_size, model_dim)
        """
        residual = query
        query = query.unsqueeze(-1)
        key = key.unsqueeze(-1)
        value = value.unsqueeze(-1)

        dim_per_head = self.dim_per_head
        num_heads = self.num_heads

        # linear projection
        key = self.linear_k(key)
        value = self.linear_v(value)
        query = self.linear_q(query)

        # split by heads
        # 没有permute，错误写法，实际上相当于始终没有把head拆分开，一直都是拖着head * d作为最后一个维度
        key = key.view(-1, num_heads, self.model_dim, dim_per_head)
        value = value.view(-1, num_heads, self.model_dim, dim_per_head)
        query = query.view(-1, num_heads, self.model_dim, dim_per_head)

        # scaled dot product attention
        # scale 没有取对，变成了 dim_per_head // num_heads
        scale = (key.size(-1) // num_heads) ** -0.5
        attention = self.dot_product_attention(query, key, value,
                                               scale)

        attention = attention.view(-1, self.model_dim, dim_per_head * num_heads)

        # final linear projection W_o
        output = self.linear_final(attention).squeeze(-1)

        # add&norm
        output = self.dropout(output)
        output = self.layer_norm(residual.squeeze(-1) + output)

        return output


class _CoAttentionLayer(nn.Module):
    """
    co-attention layer with 2 co-attention blocks
    """

    def __init__(self, model_dim=256, num_heads=8, ffn_dim=2048, dropout=0.5):
        super(_CoAttentionLayer, self).__init__()

        self.attention_1 = _MultiHeadAttention(model_dim, num_heads, dropout)
        self.ffn1 = FFN(model_dim, ffn_dim, dropout=dropout)
        self.ffn_addnorm1 = AddNorm(model_dim, dropout)

        self.attention_2 = _MultiHeadAttention(model_dim, num_heads, dropout)
        self.ffn2 = FFN(model_dim, ffn_dim, dropout=dropout)
        self.ffn_addnorm2 = AddNorm(model_dim, dropout)

        self.fusion_linear = nn.Linear(model_dim * 2, model_dim)

    def forward(self, image, text):
        """
        Args:
            image: shape=(batch_size, model_dim)
            text: shape=(batch_size, model_dim)

        Returns:
            fusion_output: shape=(batch_size, model_dim)
        """
        image = image.unsqueeze(-1)
        text = text.unsqueeze(-1)

        output1 = self.attention_1(image, text, text)
        output1 = self.ffn_addnorm1(output1, self.ffn1(output1))

        output2 = self.attention_2(text, image, image)
        output2 = self.ffn_addnorm2(output2, self.ffn2(output2))

        output = torch.cat([output1, output2], dim=1)
        output = self.fusion_linear(output)

        return output


class MCAN(AbstractModel):
    """
    Multimodal Fusion with Co-Attention Networks for Fake News Detection
    """

    def __init__(self,
                 bert: str,
                 kernel_sizes: Optional[List[int]] = None,
                 num_channels: Optional[List[int]] = None,
                 model_dim=256,
                 drop_and_bn='drop-bn',
                 num_layers=1,
                 num_heads=8,
                 ffn_dim=2048,
                 dropout=0.5):
        """
        Args:
            bert (str): bert model name
            kernel_sizes (List[int]): kernel sizes of DctCNN. Default=[3, 3, 3]
            num_channels (List[int]): number of channels of DctCNN. Default=[32, 64, 128]
            model_dim (int): model dimension. Default=256
            drop_and_bn (str): dropout and batch normalization. 'drop-bn', 'bn-drop', 'drop', 'bn' or None. Default='drop-bn'
            num_layers (int): number of co-attention layers. Default=1
            num_heads (int): number of heads in multi-head attention. Default=8
            ffn_dim (int): dimension of feed forward network. Default=2048
            dropout (float): dropout rate. Default=0.5
        """

        super(MCAN, self).__init__()

        # check input
        if kernel_sizes is None:
            kernel_sizes = [3, 3, 3]
        elif len(kernel_sizes) != 3 or not all(type(x) == int for x in kernel_sizes) or not all(
                x > 0 for x in kernel_sizes):
            raise ValueError("kernel_sizes must be a list of 3 positive integers")

        if num_channels is None:
            num_channels = [32, 64, 128]
        elif len(num_channels) != 3 or not all(type(x) == int for x in num_channels) or not all(
                x > 0 for x in num_channels):
            raise ValueError("num_channels must be a list of 3 positive integers")
        assert drop_and_bn in ['drop-bn', 'bn-drop', 'drop-only', 'bn-only', 'none'], \
            "drop_and_bn must be one of 'drop-bn', 'bn-drop', 'drop-only', 'BN-only', 'none'"

        self.model_dim = model_dim
        self.drop_and_bn = drop_and_bn

        # text
        self.bert = BertModel.from_pretrained(bert)
        self.linear_text = nn.Linear(self.bert.config.hidden_size, model_dim)
        self.bn_text = nn.BatchNorm1d(model_dim)

        # vgg image
        self.vgg = _VGG()
        self.linear_vgg = nn.Linear(4096, model_dim)
        self.bn_vgg = nn.BatchNorm1d(model_dim)

        # dct image
        self.dct_img = _DctCNN(
            dropout,
            kernel_sizes,
            num_channels,
            in_channel=128,
            branch1_channels=[64],
            branch2_channels=[48, 64],
            branch3_channels=[64, 96, 96],
            branch4_channels=[32],
            out_channels=64)
        self.linear_dct = nn.Linear(4096, model_dim)
        self.bn_dct = nn.BatchNorm1d(model_dim)

        # multimodal fusion
        self.fusion_layers = nn.ModuleList([
            _CoAttentionLayer(model_dim, num_heads, ffn_dim, dropout)
            for _ in range(num_layers)
        ])

        # classifier
        self.linear1 = nn.Linear(model_dim, 35)
        self.fc_bn = nn.BatchNorm1d(35)
        self.linear2 = nn.Linear(35, 2)
        self.dropout = nn.Dropout(dropout)

    def drop_bn_layer(self, x, part='dct'):
        bn = None
        if part == 'dct':
            bn = self.bn_dct
        elif part == 'vgg':
            bn = self.bn_vgg
        elif part == 'bert':
            bn = self.bn_text

        if self.drop_and_bn == 'drop-bn':
            x = self.dropout(x)
            x = bn(x)
        elif self.drop_and_bn == 'bn-drop':
            x = bn(x)
            x = self.dropout(x)
        elif self.drop_and_bn == 'drop':
            x = self.dropout(x)
        elif self.drop_and_bn == 'bn':
            x = bn(x)
        elif self.drop_and_bn is None:
            pass

        return x

    def forward(self, input_ids: torch.Tensor,
                mask: torch.Tensor,
                image: torch.Tensor,
                dct_img: torch.Tensor) -> torch.Tensor:
        """
        Args:
            input_ids (Tensor): shape=(batch_size, max_len)
            mask (Tensor): shape=(batch_size, max_len)
            image (Tensor): transformed image tensor, shape=(batch_size, 3, width, height)
            dct_img (Tensor): dtc image tensor, shape=(batch_size, N*N, 250)

        Returns:
              output (Tensor): shape=(batch_size, 2)
        """

        # textual feature
        bert_output = self.bert(input_ids=input_ids,
                                attention_mask=mask)
        text_output = bert_output.pooler_output
        text_output = F.relu(self.linear_text(text_output))
        text_output = self.drop_bn_layer(text_output, part='bert')

        # vgg feature
        vgg_output = self.vgg(image)
        vgg_output = F.relu(self.linear_vgg(vgg_output))
        vgg_output = self.drop_bn_layer(vgg_output, part='vgg')

        # dct feature
        dct_output = self.dct_img(dct_img)
        dct_output = F.relu(self.linear_dct(dct_output))
        dct_output = self.drop_bn_layer(dct_output, part='dct')

        output = None
        for fusion_layer in self.fusion_layers:
            output = fusion_layer(vgg_output, dct_output)

        for fusion_layer in self.fusion_layers:
            output = fusion_layer(output, text_output)

        output = F.relu(self.linear1(output))
        output = self.dropout(output)
        output = self.linear2(output)

        return output

    def calculate_loss(self, data) -> Tensor:
        token_id = data['text']['token_id']
        mask = data['text']['mask']
        vgg_feature = data['image']['vgg']
        dct_feature = data['image']['dct']
        label = data['label']
        output = self.forward(token_id, mask, vgg_feature, dct_feature)
        return F.cross_entropy(output, label)

    def predict(self, data):
        token_id = data['text']['token_id']
        mask = data['text']['mask']
        vgg_feature = data['image']['vgg']
        dct_feature = data['image']['dct']
        return torch.softmax(self.forward(token_id, mask, vgg_feature, dct_feature), dim=-1)



from typing import Optional, Callable, List

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.autograd import Function


class TextCNNLayer(nn.Module):
    """
    It's not a full TextCNN model. Only convolution and max pooling are including here
    but without an embedding layer or fully connected layer. Thus, it should be a part of your own TextCNN model
    """
    def __init__(self,
                 embedding_dim: int,
                 filter_num: int,
                 filter_sizes: List[int],
                 activate_fn: Optional[Callable] = None):
        super().__init__()
        # in_channel=1, out_channel=filter_num
        self.convs = nn.ModuleList([
            nn.Conv2d(1, filter_num, (k, embedding_dim)) for k in filter_sizes
        ])
        self.activate_fn = activate_fn

    def forward(self, embedded_text: torch.Tensor):
        # before unsqueeze: batch_size * max_len * embedding_dim
        embedded_text = embedded_text.unsqueeze(1)

        # before squeeze: batch_size * filter_num * (max_len-k+1) * 1
        if self.activate_fn is None:
            conv_features = [
                conv(embedded_text).squeeze(3) for conv in self.convs
            ]
        else:
            conv_features = [
                self.activate_fn(conv(embedded_text).squeeze(3))
                for conv in self.convs
            ]

        # before squeeze: batch_size * filter_num * 1
        # conv.shape[2] = (max_len-k+1)
        pool_features = [
            torch.max_pool1d(conv, conv.shape[2]).squeeze(2)
            for conv in conv_features
        ]

        # batch_size * (filter_num * len(filter_sizes))
        concat_features = torch.cat(pool_features, dim=1)
        return concat_features


class GradientReverseLayer(Function):
    @staticmethod
    def forward(ctx, x, lambd):
        ctx.lambd = lambd
        return x.view_as(x)

    @staticmethod
    def backward(ctx, grad_output):
        # 只需要对输入的x返回loss，其他的返回None
        # 详见 https://zhuanlan.zhihu.com/p/263827804
        return grad_output * -ctx.lambd, None


class SignedAttention(nn.Module):
    def __init__(self,
                 in_features: int,
                 out_features: int,
                 dropout: float,
                 alpha: float,
                 concat=True):
        super(SignedAttention, self).__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.dropout = dropout
        self.alpha = alpha
        self.concat = concat
        self.W = nn.Parameter(torch.zeros(size=(in_features, out_features)))
        self.a = nn.Parameter(torch.zeros(size=(2 * out_features, 1)))
        self.fc_W = nn.Parameter(
            torch.zeros(size=(2 * out_features, out_features)))
        self.leaky_relu = nn.LeakyReLU(self.alpha)

        self.__init_weights__()

    def __init_weights__(self):
        nn.init.xavier_uniform_(self.W.data, gain=1.414)
        nn.init.xavier_uniform_(self.a.data, gain=1.414)
        nn.init.xavier_uniform_(self.fc_W.data, gain=1.414)

    def forward(self, x: torch.Tensor, adj: torch.Tensor):

        h = torch.mm(x, self.W)
        Wh1 = torch.mm(h, self.a[:self.out_features, :])
        Wh2 = torch.mm(h, self.a[self.out_features:, :])
        e = self.leaky_relu(Wh1 + Wh2.T)
        zero_vec = -1e12 * torch.ones_like(e)

        attention = torch.where(adj > 0, e, zero_vec)  # [N, N]
        attention = F.softmax(attention, dim=1)
        attention = F.dropout(attention, self.dropout, training=self.training)

        negative_attention = torch.where(adj > 0, -e, zero_vec)
        negative_attention = -F.softmax(negative_attention, dim=1)
        negative_attention = F.dropout(negative_attention,
                                       self.dropout,
                                       training=self.training)
        h_prime = torch.matmul(attention, x)
        h_prime_negative = torch.matmul(negative_attention, x)
        h_prime_double = torch.cat([h_prime, h_prime_negative], dim=1)
        new_h_prime = torch.mm(h_prime_double, self.fc_W)
        if self.concat:
            return F.elu(new_h_prime)
        else:
            return new_h_prime

    def __repr__(self):
        return self.__class__.__name__ + ' (' + str(
            self.in_features) + ' -> ' + str(self.out_features) + ')'


class SignedGAT(nn.Module):
    def __init__(self,
                 node_vectors: torch.Tensor,
                 cos_sim_matrix: torch.Tensor,
                 num_features: int,
                 node_num: int,
                 adj_matrix: torch.Tensor,
                 head_num=4,
                 out_features=300,
                 dropout=0,
                 alpha=0.3):
        super(SignedGAT, self).__init__()
        self.dropout = dropout
        self.node_num = node_num
        self.node_embedding = nn.Embedding.from_pretrained(node_vectors, padding_idx=0)
        self.original_adj = adj_matrix
        self.potential_adj = torch.where(cos_sim_matrix > 0.5,
                                         torch.ones_like(cos_sim_matrix),
                                         torch.zeros_like(cos_sim_matrix))
        self.adj = self.original_adj + self.potential_adj
        self.adj = torch.where(self.adj > 0, torch.ones_like(self.adj),
                               torch.zeros_like(self.adj))

        self.attentions = [
            SignedAttention(num_features,
                            out_features,
                            dropout=dropout,
                            alpha=alpha,
                            concat=True) for _ in range(head_num)
        ]
        for i, attention in enumerate(self.attentions):
            self.add_module('attention_{}'.format(i), attention)

        # multi head attention
        self.out_att = SignedAttention(num_features * head_num,
                                       out_features,
                                       dropout=dropout,
                                       alpha=alpha,
                                       concat=False)

    def forward(self, post_id: torch.Tensor):
        embedding = self.node_embedding(torch.arange(
            0, self.node_num).long()).to(torch.float32)
        x = F.dropout(embedding, self.dropout, training=self.training)
        adj = self.adj.to(torch.float32)
        x = torch.cat([att(x, adj) for att in self.attentions], dim=1)
        x = F.dropout(x, self.dropout, training=self.training)
        x = torch.sigmoid(self.out_att(x, adj))
        return x[post_id]



from faknow.run.content_based import *
from faknow.run.social_context import *
from faknow.run.knowledge_aware import *

__all__ = ['run', 'run_from_yaml']


def run(model: str, **kwargs):
    model = model.lower()
    eval(f'run_{model}(**{kwargs})')


def run_from_yaml(model: str, path: str):
    model = model.lower()
    eval(f"run_{model}_from_yaml('{path}')")



from faknow.run.content_based.multimodal.run_eann import run_eann, run_eann_from_yaml, TokenizerEANN, transform_eann, \
    adjust_lr_eann
from faknow.run.content_based.multimodal.run_mcan import run_mcan, run_mcan_from_yaml, TokenizerMCAN, transform_mcan, \
    process_dct_mcan, get_optimizer_mcan
from faknow.run.content_based.multimodal.run_mfan import run_mfan, run_mfan_from_yaml, TokenizerMFAN, transform_mfan, \
    load_adj_matrix_mfan
from faknow.run.content_based.multimodal.run_safe import run_safe, run_safe_from_yaml
from faknow.run.content_based.multimodal.run_spotfake import run_spotfake, run_spotfake_from_yaml, TokenizerSpotFake, \
    text_preprocessing, transform_spotfake



from typing import Optional, Tuple, Union

import torch
from torch import nn


def conv1d_bn_relu(in_channels, out_channels, kernel_size, stride=1, padding=0):
    return nn.Sequential(
        nn.Conv1d(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
        ),
        nn.BatchNorm1d(out_channels),
        nn.ReLU(inplace=True),
    )


def conv2d_bn_relu(in_channels, out_channels, kernel_size, stride=1, padding: Union[str, int, Tuple[int, int]] = 0):
    return nn.Sequential(
        nn.Conv2d(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=(1, kernel_size),
            stride=stride,
            padding=padding,
        ),
        nn.BatchNorm2d(out_channels),
        nn.ReLU(inplace=True),
    )


class DctStem(nn.Module):
    def __init__(self, kernel_sizes, num_channels):
        super(DctStem, self).__init__()
        self.convs = nn.Sequential(
            conv2d_bn_relu(in_channels=1,
                           out_channels=num_channels[0],
                           kernel_size=kernel_sizes[0]),
            conv2d_bn_relu(
                in_channels=num_channels[0],
                out_channels=num_channels[1],
                kernel_size=kernel_sizes[1],
            ),
            conv2d_bn_relu(
                in_channels=num_channels[1],
                out_channels=num_channels[2],
                kernel_size=kernel_sizes[2],
            ),
            nn.MaxPool2d((1, 2)),
        )

    def forward(self, dct_img):
        x = dct_img.unsqueeze(1)
        img = self.convs(x)
        img = img.permute(0, 2, 1, 3)

        return img


class DctInceptionBlock(nn.Module):
    def __init__(
            self,
            in_channel=128,
            branch1_channels=None,
            branch2_channels=None,
            branch3_channels=None,
            branch4_channels=None,
    ):
        super(DctInceptionBlock, self).__init__()

        if branch4_channels is None:
            branch4_channels = [32]
        if branch3_channels is None:
            branch3_channels = [64, 96, 96]
        if branch2_channels is None:
            branch2_channels = [48, 64]
        if branch1_channels is None:
            branch1_channels = [64]

        self.branch1 = conv2d_bn_relu(in_channels=in_channel,
                                      out_channels=branch1_channels[0],
                                      kernel_size=1)

        self.branch2 = nn.Sequential(
            conv2d_bn_relu(in_channels=in_channel,
                           out_channels=branch2_channels[0],
                           kernel_size=1),
            conv2d_bn_relu(
                in_channels=branch2_channels[0],
                out_channels=branch2_channels[1],
                kernel_size=3,
                padding=(0, 1),
            ),
        )

        self.branch3 = nn.Sequential(
            conv2d_bn_relu(in_channels=in_channel,
                           out_channels=branch3_channels[0],
                           kernel_size=1),
            conv2d_bn_relu(
                in_channels=branch3_channels[0],
                out_channels=branch3_channels[1],
                kernel_size=3,
                padding=(0, 1),
            ),
            conv2d_bn_relu(
                in_channels=branch3_channels[1],
                out_channels=branch3_channels[2],
                kernel_size=3,
                padding=(0, 1),
            ),
        )

        self.branch4 = nn.Sequential(
            nn.MaxPool2d(kernel_size=(1, 3), stride=1, padding=(0, 1)),
            conv2d_bn_relu(in_channels=in_channel,
                           out_channels=branch4_channels[0],
                           kernel_size=1),
        )

    def forward(self, x):
        x = x.permute(0, 2, 1, 3)
        out1 = self.branch1(x)
        out2 = self.branch2(x)
        out3 = self.branch3(x)
        out4 = self.branch4(x)
        out = torch.cat([out1, out2, out3, out4], dim=1)
        out = out.permute(0, 2, 1, 3)

        return out



import os

import torch.utils.data
from torch.utils.data import ConcatDataset, random_split
from torchvision.datasets.folder import find_classes, is_image_file
from typing import Optional, Callable, Dict, List, Tuple, Union, Set


def default_walker(class_dir: str, class_name: str, class_index: int,
                   instances: List[Tuple], available_classes: Set):
    for root, _, file_names in sorted(os.walk(class_dir, followlinks=True)):
        for file_name in sorted(file_names):
            path = os.path.join(root, file_name)
            instances.append((path, class_index))

            if class_name not in available_classes:
                available_classes.add(class_name)


def walker_with_images(class_dir: str, class_name: str, class_index: int,
                       instances: List[Tuple], available_classes: Set):
    for root, _, file_names in sorted(os.walk(class_dir, followlinks=True)):
        find_text = find_image = False
        for file_name in sorted(file_names):
            if is_image_file(file_name):
                image_path = os.path.join(root, file_name)
                find_image = True
            else:
                text_path = os.path.join(root, file_name)
                find_text = True

            if find_text and find_image:
                instances.append((text_path, image_path, class_index))
                find_text = find_image = False

            if class_name not in available_classes:
                available_classes.add(class_name)


def make_dataset(
    directory: str,
    class_to_idx: Optional[Dict[str, int]] = None,
    extensions: Optional[Union[str, Tuple[str, ...]]] = None,
    is_valid_file: Optional[Callable[[str], bool]] = None,
    walk_class_dir: Optional[Callable[[str, str, int, List[Tuple], Set],
                                      None]] = default_walker
) -> List[Tuple[str, int]]:
    directory = os.path.expanduser(directory)

    if class_to_idx is None:
        _, class_to_idx = find_classes(directory)
    elif not class_to_idx:
        raise ValueError(
            "'class_to_index' must have at least one entry to collect any samples."
        )

    # both_none = extensions is None and is_valid_file is None
    # both_something = extensions is not None and is_valid_file is not None
    # if both_none or both_something:
    #     raise ValueError("Both extensions and is_valid_file cannot be None or not None at the same time")
    #
    # if extensions is not None:
    #     def is_valid_file(x: str) -> bool:
    #         return has_file_allowed_extension(x, extensions)
    #
    # is_valid_file = cast(Callable[[str], bool], is_valid_file)

    instances = []
    available_classes = set()
    for class_name in sorted(class_to_idx.keys()):
        class_index = class_to_idx[class_name]
        class_dir = os.path.join(directory, class_name)
        if not os.path.isdir(class_dir):
            continue

        walk_class_dir(class_dir, class_name, class_index, instances,
                       available_classes)

    empty_classes = set(class_to_idx.keys()) - available_classes
    if empty_classes:
        msg = f"Found no valid file for the classes {', '.join(sorted(empty_classes))}. "
        if extensions is not None:
            msg += f"Supported extensions are: {extensions if isinstance(extensions, str) else ', '.join(extensions)}"
        raise FileNotFoundError(msg)

    return instances


def re_split_dataset(datasets: List[torch.utils.data.Dataset],
                     lengths: List[int]):
    all_dataset = ConcatDataset(datasets)
    return random_split(all_dataset, lengths)



import torch
from torch_geometric.loader import DataLoader

from faknow.train.trainer import BaseTrainer


class BaseGNNTrainer(BaseTrainer):

    @torch.no_grad()
    def evaluate(self, loader: DataLoader):
        self.model.eval()
        outputs = []
        labels = []
        for batch_data in loader:
            outputs.append(self.model.predict(batch_data))
            labels.append(batch_data.y)
        return self.evaluator.evaluate(torch.concat(outputs),
                                       torch.concat(labels))



from torch.utils.data import Dataset


class NEPDataset(Dataset):
    def __init__(self, post_simcse, avg_mac, avg_mic, p_mac, p_mic, avg_mic_mic, token, label):
        super().__init__()
        self.post_simcse = post_simcse
        self.avg_mac = avg_mac
        self.avg_mic = avg_mic
        self.p_mac = p_mac
        self.p_mic = p_mic
        self.avg_mic_mic = avg_mic_mic
        self.token = token
        self.label = label

    def __getitem__(self, index):
        item = {
            'post_simcse': self.post_simcse[index],
            'avg_mac': self.avg_mac[index],
            'avg_mic': self.avg_mic[index],
            'kernel_p_mac': self.p_mac[index],
            'kernel_p_mic': self.p_mic[index],
            'kernel_avg_mic_mic': self.avg_mic_mic[index],
            'token': self.token[index],
            'label': self.label[index]
        }
        return item

    def __len__(self):
        return len(self.label)



import torch.utils.data
from torch.utils.data.dataset import T_co


# including data processing, downloading some open datasets and some custom configuration
class Dataset(torch.utils.data.Dataset):
    def __init__(self) -> None:
        super().__init__()

    def __str__(self) -> str:
        return super().__str__()

    def __getitem__(self, index) -> T_co:
        pass



from pathlib import Path
from typing import Optional, Callable, Dict, Tuple, Any

import numpy as np
import torch.utils.data
from nltk.tokenize import sent_tokenize

from faknow.data.legacy.text_dataset import FolderTextDataset


class SAFENumpyDataset(torch.utils.data.Dataset):
    def __init__(
            self,
            root_dir: str,
    ):
        super().__init__()

        x_heads = np.load(root_dir + "\\case_headline.npy", allow_pickle=True)
        x_bodies = np.load(root_dir + "\\case_body.npy", allow_pickle=True)
        x_images = np.load(root_dir + "\\case_image.npy", allow_pickle=True)
        y = np.load(root_dir + "\\case_y_fn_dim1.npy")

        self.x_heads = x_heads.astype(np.float32)
        self.x_bodies = x_bodies.astype(np.float32)
        self.x_images = x_images.astype(np.float32)
        self.y = y.astype(np.float32)

        assert self.x_heads.shape[0] == self.x_bodies.shape[0] == self.x_images.shape[0] == self.y.shape[0]

    def __len__(self):
        return self.x_heads.shape[0]

    def __getitem__(self, index: int):
        return {'head': self.x_heads[index], 'body': self.x_bodies[index], 'image': self.x_images[index], 'label': self.y[index]}


class SAFEDataset(FolderTextDataset):
    def __init__(
            self,
            root: str,
            embedding: Callable[[str, Optional[Dict]], Any],
            max_len: int = 16,
    ):
        super().__init__(root, embedding=embedding)

        self.max_len = max_len

    def _embedding(self, text: str) -> np.ndarray:
        sentences = sent_tokenize(text)
        embd = self.embedding(sentences)
        return embd

    def _padding(self, x: np.ndarray) -> np.ndarray:
        if x.shape[0] >= self.max_len:
            return x[:self.max_len]
        else:
            padded = np.zeros((self.max_len, *x.shape[1:]), dtype=x.dtype)
            padded[:x.shape[0]] = x
            return padded

    def __getitem__(self, index) -> Tuple:
        text_path, label = self.samples[index]
        with open(text_path, encoding='utf-8') as f:
            text = f.read()
        headline, body, image = text.split("\n")

        headline = self._padding(self._embedding(headline.strip())).astype(np.float32)
        body = self._padding(self._embedding(body.strip())).astype(np.float32)
        image = self._padding(self._embedding(image.strip())).astype(np.float32)

        return headline, body, image, label



from typing import Optional, Union, Iterable, Sequence

import torch.utils.data
from torch.utils.data import Dataset, Sampler
from torch.utils.data.dataloader import T_co, _collate_fn_t, _worker_init_fn_t, _BaseDataLoaderIter


# you can use custom sampler here
class DataLoader(torch.utils.data.DataLoader):
    def __init__(self, dataset: Dataset[T_co], batch_size: Optional[int] = 1, shuffle: Optional[bool] = None,
                 sampler: Union[Sampler, Iterable, None] = None,
                 batch_sampler: Union[Sampler[Sequence], Iterable[Sequence], None] = None, num_workers: int = 0,
                 collate_fn: Optional[_collate_fn_t] = None, pin_memory: bool = False, drop_last: bool = False,
                 timeout: float = 0, worker_init_fn: Optional[_worker_init_fn_t] = None, multiprocessing_context=None,
                 generator=None, *, prefetch_factor: int = 2, persistent_workers: bool = False,
                 pin_memory_device: str = ""):
        super().__init__(dataset, batch_size, shuffle, sampler, batch_sampler, num_workers, collate_fn, pin_memory,
                         drop_last, timeout, worker_init_fn, multiprocessing_context, generator,
                         prefetch_factor=prefetch_factor, persistent_workers=persistent_workers,
                         pin_memory_device=pin_memory_device)

    def __iter__(self) -> '_BaseDataLoaderIter':
        return super().__iter__()



from faknow.run import run, run_from_yaml
from faknow.model.content_based.multi_modal.eann import EANN
from faknow.run.content_based.multimodal.run_eann import run_eann, run_eann_from_yaml


if __name__ == '__main__':
    model = 'eann'
    run_from_yaml(model, r'..\properties\eann.yaml')
    # model = 'mdfend'
    # run(model, train_path=r'F:\dataset\weibo21-mdfend\all.json')



import torch
import torch.nn.functional as F
from torch import Tensor, nn
from torch_geometric.nn import GATConv, GCNConv, SAGEConv, global_max_pool

from faknow.model.model import AbstractModel


class _BaseGNN(AbstractModel):
    def __init__(self, feature_size: int, hidden_size: int, concat=False):
        """

        Args:
            feature_size (int): dimension of input node feature
            hidden_size (int): Default=128
            concat (bool): concat news embedding and graph embedding. Default=False
        """
        super(_BaseGNN, self).__init__()
        self.feature_size = feature_size
        self.hidden_size = hidden_size
        self.concat = concat

        if self.concat:
            self.fc0 = torch.nn.Linear(self.feature_size, self.hidden_size)
            self.fc1 = torch.nn.Linear(self.hidden_size * 2, self.hidden_size)
        self.fc2 = torch.nn.Linear(self.hidden_size, 2)

    def forward(self, x: Tensor, edge_index: Tensor, batch: Tensor,
                num_graphs: int):
        """

        Args:
            x (Tensor): node feature, shape=(num_nodes, feature_size)
            edge_index (Tensor): edge index, shape=(2, num_edges)
            batch (Tensor): index of graph each node belongs to, shape=(num_nodes,)
            num_graphs (int): number of graphs, a.k.a. batch_size

        Returns:
            output (Tensor): prediction of being fake, shape=(num_graphs, 2)
        """
        edge_attr = None
        raw_x = x
        x = F.relu(self.conv(x, edge_index, edge_attr))
        x = global_max_pool(x, batch)

        # whether concat news embedding and graph embedding
        if self.concat:
            news = torch.stack([
                raw_x[(batch == idx).nonzero().squeeze()[0]]
                for idx in range(num_graphs)
            ])
            news = F.relu(self.fc0(news))
            x = torch.cat([x, news], dim=1)
            x = F.relu(self.fc1(x))
        x = self.fc2(x)
        return x

    def calculate_loss(self, data) -> torch.Tensor:
        output = self.forward(data.x, data.edge_index, data.batch,
                              data.num_graphs)
        loss_fn = nn.CrossEntropyLoss()
        loss = loss_fn(output, data.y)
        return loss

    def predict(self, data_without_label) -> torch.Tensor:
        output = self.forward(data_without_label.x,
                              data_without_label.edge_index,
                              data_without_label.batch,
                              data_without_label.num_graphs)
        return F.softmax(output, dim=1)


class GCN(_BaseGNN):
    """
    Semi-Supervised Classification with Graph Convolutional Networks
    paper: https://ieeexplore.ieee.org/document/8953909
    code: https://github.com/safe-graph/GNN-FakeNews
    """
    def __init__(self, feature_size: int, hidden_size=128):
        """

        Args:
            feature_size (int): dimension of input node feature
            hidden_size (int): Default=128
        """
        super().__init__(feature_size, hidden_size, False)
        self.conv = GCNConv(self.feature_size, self.hidden_size)


class SAGE(_BaseGNN):
    """
    Inductive Representation Learning on Large Graphs
    paper: https://dl.acm.org/doi/10.5555/3294771.3294869
    code: https://github.com/safe-graph/GNN-FakeNews
    """
    def __init__(self, feature_size: int, hidden_size=128):
        """

        Args:
            feature_size (int): dimension of input node feature
            hidden_size (int): Default=128
        """
        super().__init__(feature_size, hidden_size, False)
        self.conv = SAGEConv(self.feature_size, self.hidden_size)


class GAT(_BaseGNN):
    """
    Graph Attention Networks
    paper: https://openreview.net/forum?id=rJXMpikCZ
    code: https://github.com/safe-graph/GNN-FakeNews
    """
    def __init__(self, feature_size: int, hidden_size=128):
        """

        Args:
            feature_size (int): dimension of input node feature
            hidden_size (int): Default=128
        """
        super().__init__(feature_size, hidden_size, False)
        self.conv = GATConv(self.feature_size, self.hidden_size)



import torch
import torch.nn.functional as F
from torch import nn, Tensor
from torch_geometric.nn import GATConv, global_mean_pool

from faknow.model.model import AbstractModel

"""
Fake news detection on social media using geometric deep learning
paper: https://arxiv.org/abs/1902.06673
code: https://github.com/safe-graph/GNN-FakeNews
"""


class _BaseGCNFN(AbstractModel):
    def __init__(self,
                 feature_size: int,
                 hidden_size=128,
                 dropout_ratio=0.5,
                 concat=False):
        """

        Args:
            feature_size (int): dimension of input node feature
            hidden_size (int): dimension of hidden layer. Default=128
            dropout_ratio (float): dropout ratio. Default=0.5
            concat (bool): concat news embedding and graph embedding. Default=False
        """
        super(_BaseGCNFN, self).__init__()

        self.feature_size = feature_size
        self.hidden_size = hidden_size
        self.dropout_ratio = dropout_ratio
        self.concat = concat

        self.conv1 = GATConv(self.feature_size, self.hidden_size * 2)
        self.conv2 = GATConv(self.hidden_size * 2, self.hidden_size * 2)

        if self.concat:
            self.fc0 = nn.Linear(self.feature_size, self.hidden_size)

        self.fc1 = nn.Linear(self.hidden_size * 2, self.hidden_size)
        self.fc2 = nn.Linear(self.hidden_size, 2)

    def forward(self, x: Tensor, edge_index: Tensor, batch: Tensor,
                num_graphs: int):
        """

        Args:
            x (Tensor): node feature, shape=(num_nodes, feature_size)
            edge_index (Tensor): edge index, shape=(2, num_edges)
            batch (Tensor): index of graph each node belongs to, shape=(num_nodes,)
            num_graphs (int): number of graphs, a.k.a. batch_size

        Returns:
            output (Tensor): prediction of being fake, shape=(num_graphs, 2)
        """
        raw_x = x
        x = F.selu(self.conv1(x, edge_index))
        x = F.selu(self.conv2(x, edge_index))
        x = F.selu(global_mean_pool(x, batch))
        x = F.selu(self.fc1(x))
        x = F.dropout(x, p=self.dropout_ratio, training=self.training)

        # whether concat news embedding and graph embedding
        if self.concat:
            news = torch.stack([
                raw_x[(batch == idx).nonzero().squeeze()[0]]
                for idx in range(num_graphs)
            ])
            news = F.relu(self.fc0(news))
            x = torch.cat([x, news], dim=1)
            x = F.relu(self.fc1(x))

        out = self.fc2(x)
        return out

    def calculate_loss(self, data) -> torch.Tensor:
        output = self.forward(data.x, data.edge_index, data.batch,
                              data.num_graphs)
        loss_fn = nn.CrossEntropyLoss()
        loss = loss_fn(output, data.y)
        return loss

    def predict(self, data_without_label) -> torch.Tensor:
        output = self.forward(data_without_label.x,
                              data_without_label.edge_index,
                              data_without_label.batch,
                              data_without_label.num_graphs)
        return F.softmax(output, dim=1)


class GCNFN(_BaseGCNFN):
    """
    Fake news detection on social media using geometric deep learning
    """
    def __init__(self,
                 feature_size: int,
                 hidden_size=128,
                 dropout_ratio=0.5):
        """

        Args:
            feature_size (int): dimension of input node feature
            hidden_size (int): dimension of hidden layer. Default=128
            dropout_ratio (float): dropout ratio. Default=0.5
        """
        super(GCNFN, self).__init__(feature_size, hidden_size, dropout_ratio, False)


import logging
import os
import warnings
from time import time
from typing import Optional, Dict, Union, Any, Tuple

import torch
from torch.optim.lr_scheduler import _LRScheduler
from torch.optim.optimizer import Optimizer
from torch.utils.data.dataloader import DataLoader
from torch.utils.tensorboard import SummaryWriter
from tqdm import tqdm

from faknow.evaluate.evaluator import Evaluator
from faknow.model.model import AbstractModel
from faknow.utils.util import dict2str, seconds2str, now2str, check_loss_type, EarlyStopping


class AbstractTrainer:
    def __init__(self,
                 model: AbstractModel,
                 evaluator: Evaluator,
                 optimizer: Optimizer,
                 scheduler: Optional[_LRScheduler] = None,
                 clip_grad_norm: Optional[Dict[str, Any]] = None,
                 device='cpu'):
        self.model = model
        self.optimizer = optimizer
        self.evaluator = evaluator
        self.scheduler = scheduler
        self.clip_grad_norm = clip_grad_norm
        self.device = torch.device(device)
        self.model.to(self.device)

    @torch.no_grad()
    def evaluate(self, data: DataLoader):
        """evaluate after training"""
        raise NotImplementedError

    def fit(self,
            train_loader: DataLoader,
            num_epochs: int,
            validate_loader: Optional[DataLoader] = None,
            save=False,
            save_path=None):
        raise NotImplementedError

    def to(self, device: str, **kwargs):
        self.device = torch.device(device)
        self.model.to(self.device, **kwargs)

    def cuda(self, device='cuda:0', **kwargs):
        self.to(device, **kwargs)

    def cpu(self, **kwargs):
        self.to('cpu', **kwargs)

    def _move_data_to_device(self, batch_data) -> Any:
        if type(batch_data) is dict:
            for k, v in batch_data.items():
                if type(v) is torch.Tensor:
                    batch_data[k] = v.to(self.device)
                else:
                    # todo 递归字典的情况
                    batch_data[k] = self._move_data_to_device(v)
        elif type(batch_data) is tuple:
            batch_data = tuple(value.to(self.device) for value in batch_data)
        else:
            batch_data = batch_data.to(self.device)
        return batch_data


class BaseTrainer(AbstractTrainer):
    def __init__(self,
                 model: AbstractModel,
                 evaluator: Evaluator,
                 optimizer: Optimizer,
                 scheduler: Optional[_LRScheduler] = None,
                 clip_grad_norm: Optional[Dict[str, Any]] = None,
                 device='cpu',
                 early_stopping: Optional[EarlyStopping] = None):
        super(BaseTrainer, self).__init__(model, evaluator, optimizer,
                                          scheduler, clip_grad_norm, device)

        # best validation score
        self.early_stopping = early_stopping
        self.best_score = 0.0
        self.best_epoch = 0

        self.writer = None
        self.logger = logging.getLogger(__name__)
        self.logger.setLevel(logging.DEBUG)
        formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s')
        sh = logging.StreamHandler()
        sh.setFormatter(formatter)
        self.logger.addHandler(sh)

    def _train_epoch(self, loader: DataLoader,
                     epoch: int) -> Union[float, Dict[str, float]]:
        """training for one epoch"""

        # switch model to train mode
        self.model.train()

        with tqdm(enumerate(loader),
                  total=len(loader),
                  ncols=100,
                  desc='Training') as pbar:
            loss = None
            result_is_dict = False
            for batch_id, batch_data in pbar:
                batch_data = self._move_data_to_device(batch_data)
                result = self.model.calculate_loss(batch_data)

                # check result type
                loss, result_is_dict = check_loss_type(result)

                # backward
                self.optimizer.zero_grad()
                loss.backward()

                # gradient clipping
                if self.clip_grad_norm is not None:
                    torch.nn.utils.clip_grad_norm_(
                        parameters=self.model.parameters(),
                        **self.clip_grad_norm)
                self.optimizer.step()

            if result_is_dict:
                return {k: v.item() for k, v in result.items()}
            return loss.item()

    def _validate_epoch(self, loader: DataLoader,
                        epoch: int) -> Tuple[float, Dict[str, float]]:
        """validation after training for one epoch"""
        # evaluate
        result = self.evaluate(loader)

        # get the first metric as the validation score if accuracy is not in result
        if 'accuracy' in result:
            score = result['accuracy']
        else:
            if epoch == 0:
                warnings.warn(
                    'no accuracy in result, use the first metric as the validation score'
                )
            score = list(result.values())[0]
        return score, result

    @torch.no_grad()
    def evaluate(self, loader: DataLoader):
        self.model.eval()

        outputs = []
        labels = []
        for batch_data in loader:
            batch_data = self._move_data_to_device(batch_data)
            outputs.append(self.model.predict(batch_data))

            # todo 统一使用dict还是tuple 是否要区分dict trainer和tuple trainer
            labels.append(batch_data['label'])
        return self.evaluator.evaluate(torch.concat(outputs),
                                       torch.concat(labels))

    def fit(self,
            train_loader: DataLoader,
            num_epochs: int,
            validate_loader: Optional[DataLoader] = None,
            save_best: Optional[bool] = None,
            save_path: Optional[str] = None):

        result_file_name = f"{self.model.__class__.__name__}-{now2str()}"

        # 未指定保存路径时，取fit开始时刻作为文件名
        if save_best is not None and save_path is None:
            save_path = os.path.join(os.getcwd(), f"save/{result_file_name}.pth")

        # log
        tb_logs_path = f"tb_logs/{result_file_name}"
        self.writer = SummaryWriter(tb_logs_path)
        self.__add_file_log(result_file_name)

        self.logger.info(f'Tensorboard log is saved in {tb_logs_path}')
        self.logger.info(f'log file is saved in logs/{tb_logs_path}.log\n')
        self._show_data_size(train_loader, validate_loader)
        self.logger.info('----start training-----\n')

        for epoch in range(num_epochs):
            self.logger.info(f'epoch=[{epoch}/{num_epochs - 1}]')

            # train
            training_start_time = time()
            train_result = self._train_epoch(train_loader, epoch)

            # show training result
            cost_time_str = seconds2str(time() - training_start_time)
            self._show_train_result(train_result, cost_time_str, epoch)

            # validate
            if validate_loader is not None:
                validation_score, validation_result = self._validate_epoch(
                    validate_loader, epoch)

                save_best = self._find_best_score(epoch, validation_score, save_best, save_path)

                self._show_validation_result(validation_result,
                                             validation_score,
                                             epoch,
                                             save_best)

                if self.early_stopping is not None and self.early_stopping.early_stop:
                    break

            # learning rate scheduler
            if self.scheduler is not None:
                self.scheduler.step()

        self.writer.flush()
        self.writer.close()

        # save last epoch model
        # do not use bool condition like `if not save_best`
        if save_best is False:
            self.save(save_path)

    def save(self, save_path: Optional[str] = None):

        # default save path: './save/model_name-current_time.pth'
        if save_path is None:
            save_dir = os.path.join(os.getcwd(), "save")
            file_name = f"{self.model.__class__.__name__}-{now2str()}.pth"
            save_path = os.path.join(save_dir, file_name)

        # create save dir if not exists
        if not os.path.exists(os.path.dirname(save_path)):
            os.makedirs(os.path.dirname(save_path))

        torch.save(self.model.state_dict(), save_path)

        self.logger.info(f'\nmodel is saved in {save_path}')

    def _find_best_score(self, epoch: int, validation_score: float, save_best: bool, save_path: str) -> bool:

        improvement = False
        if self.early_stopping is not None:
            save_best = True
            improvement = self.early_stopping(validation_score)
        elif save_best:
            improvement = validation_score > self.best_score

        # save best model
        if save_best and improvement:
            self.best_score = validation_score
            self.best_epoch = epoch
            self.save(save_path)

        return save_best

    def _show_train_result(self,
                           train_result: Union[float, Dict[str, float]],
                           cost_time_str: str,
                           epoch: int):

        self.logger.info(f'training time={cost_time_str}')

        if type(train_result) is float:
            # single loss
            self.writer.add_scalar("Train/loss", train_result, epoch)
            self.logger.info(f"training loss : loss={train_result:.6f}")
        elif type(train_result) is dict:
            # multiple losses
            for metric, value in train_result.items():
                self.writer.add_scalar("Train/" + metric, value, epoch)
            self.logger.info(f"training loss : {dict2str(train_result)}")
        else:
            raise TypeError(
                f"train_result type error: must be float or Dict[str, float], but got {type(train_result)}"
            )

    def _show_validation_result(self,
                                validation_result: Dict[str, float],
                                validation_score: float,
                                epoch: int,
                                save_best: Optional[bool] = None):
        for metric, value in validation_result.items():
            self.writer.add_scalar("Validation/" + metric, value, epoch)
        self.logger.info("validation result : " + dict2str(validation_result))

        score_info = f"current score : {validation_score:.6f}\n"
        if save_best:
            score_info = score_info + f", best score : {self.best_score:.6f}, best epoch : {str(self.best_epoch)}"
        if self.early_stopping is not None and self.early_stopping.early_stop:
            score_info = score_info + f"\nearly stopping at epoch {epoch}!"
        self.logger.info(score_info)

    def _show_data_size(self,
                        train_loader: DataLoader,
                        validate_loader: Optional[DataLoader] = None):
        train_set_size = len(train_loader.dataset)
        self.logger.info(f'training data size={train_set_size}')
        if validate_loader is not None:
            validate_set_size = len(validate_loader.dataset)
            self.logger.info(f'validation data size={validate_set_size}')

    def __add_file_log(self, file_name: str):
        logs_dir = os.path.join(os.getcwd(), "logs")
        if not os.path.exists(logs_dir):
            os.makedirs(logs_dir)
        log_file_path = os.path.join(logs_dir, f"{file_name}.log")

        fh = logging.FileHandler(log_file_path)
        formatter = logging.Formatter(
            '%(asctime)s - %(levelname)s - %(message)s')
        fh.setFormatter(formatter)
        self.logger.addHandler(fh)



import pickle
import re
from typing import Dict, List, Any, Optional, Callable

import jieba
import torch
import torch.nn.functional as F
import yaml
from torch.utils.data import DataLoader

from faknow.data.dataset.text import TextDataset
from faknow.evaluate.evaluator import Evaluator
from faknow.model.content_based.textcnn import TextCNN
from faknow.train.trainer import BaseTrainer
from faknow.utils.util import dict2str, read_stop_words

__all__ = ['TokenizerTextCNN', 'run_textcnn', 'run_textcnn_from_yaml']


class TokenizerTextCNN:
    def __init__(self, vocab: Dict[str, int], max_len=255, stop_words: List[str] = None, language='zh') -> None:
        assert language in ['zh', 'en'], "language must be one of {zh, en}"
        self.language = language
        self.vocab = vocab
        self.max_len = max_len
        if stop_words is None:
            stop_words = []
        self.stop_words = stop_words

    def __call__(self, texts: List[str]) -> torch.Tensor:
        token_ids = []
        for text in texts:
            cleaned_text = re.sub(u"[，。 :,.；|-“”——_/nbsp+&;@、《》～（）())#O！：【】]", "", text).strip().lower()

            if self.language == 'zh':
                split_words = jieba.cut(cleaned_text)
            else:
                split_words = cleaned_text.split()
            words = [word for word in split_words if word not in self.stop_words]

            token_id = [self.vocab[word] for word in words]
            real_len = len(token_id)
            if real_len < self.max_len:
                token_id.extend([0] * (self.max_len - real_len))
            elif real_len > self.max_len:
                token_id = token_id[:self.max_len]
            token_ids.append(token_id)

        return torch.tensor(token_ids)


def run_textcnn(train_path: str,
                vocab: Dict[str, int],
                stop_words: List[str],
                word_vectors: torch.Tensor,
                language='zh',
                max_len=255,
                filter_num=100,
                kernel_sizes: List[int] = None,
                activate_func: Optional[Callable] = F.relu,
                dropout=0.5,
                freeze=False,
                batch_size=50,
                lr=0.001,
                num_epochs=25,
                metrics: List = None,
                validate_path: str = None,
                test_path: str = None,
                device='cpu') -> None:
    tokenizer = TokenizerTextCNN(vocab, max_len, stop_words, language)

    train_set = TextDataset(train_path, ['text'], tokenizer)
    train_loader = DataLoader(train_set, batch_size, shuffle=True)

    if validate_path is not None:
        val_set = TextDataset(validate_path, ['text'], tokenizer)
        val_loader = DataLoader(val_set, batch_size, shuffle=False)
    else:
        val_loader = None

    model = TextCNN(word_vectors, filter_num, kernel_sizes, activate_func, dropout, freeze)
    optimizer = torch.optim.Adam(filter(lambda p: p.requires_grad,
                                        list(model.parameters())),
                                 lr)
    evaluator = Evaluator(metrics)
    trainer = BaseTrainer(model, evaluator, optimizer, device=device)
    trainer.fit(train_loader, num_epochs, validate_loader=val_loader)

    if test_path is not None:
        test_set = TextDataset(test_path, ['text'], tokenizer, ['image'])
        test_loader = DataLoader(test_set, batch_size, shuffle=False)
        test_result = trainer.evaluate(test_loader)
        print(f"test result: {dict2str(test_result)}")


def _parse_kargs(config: Dict[str, Any]) -> Dict[str, Any]:
    with open(config['vocab'], 'rb') as f:
        config['vocab'] = pickle.load(f)
    with open(config['word_vectors'], 'rb') as f:
        config['word_vectors'] = pickle.load(f)
    config['stop_words'] = read_stop_words(config['stop_words'])
    return config


def run_textcnn_from_yaml(path: str):
    with open(path, 'r', encoding='utf-8') as _f:
        _config = yaml.load(_f, Loader=yaml.FullLoader)
        run_textcnn(**_parse_kargs(_config))



from typing import Dict

import torch
import torch.nn as nn
from torch import Tensor
from torchvision import models

from faknow.model.layers.layer import SignedGAT, TextCNNLayer
from faknow.model.layers.transformer import (FFN,
                                             ScaledDotProductAttention, transpose_qkv, transpose_output)
from faknow.model.model import AbstractModel
from faknow.utils.util import calculate_cos_matrix

"""
MFAN: Multi-modal Feature-enhanced TransformerBlock Networks for Rumor Detection
paper: https://www.ijcai.org/proceedings/2022/335
code: https://github.com/drivsaf/MFAN
"""


class TransformerBlock(nn.Module):
    def __init__(self,
                 input_size: int,
                 key_size=16,
                 value_size=16,
                 head_num=8,
                 dropout=0.1):
        super(TransformerBlock, self).__init__()
        self.head_num = head_num
        self.k_size = key_size if key_size is not None else input_size
        self.v_size = value_size if value_size is not None else input_size

        # only for self-attention, the input dimensions of Q, K, V are the same
        self.W_q = nn.Parameter(torch.Tensor(input_size, head_num * key_size))
        self.W_k = nn.Parameter(torch.Tensor(input_size, head_num * key_size))
        self.W_v = nn.Parameter(torch.Tensor(input_size,
                                             head_num * value_size))

        self.W_o = nn.Parameter(torch.Tensor(value_size * head_num,
                                             input_size))

        self.dropout = nn.Dropout(dropout)

        self.ffn = FFN(input_size, input_size, input_size, dropout)
        self.dot_product_attention = ScaledDotProductAttention(
            epsilon=1e-6, dropout=dropout)
        self.__init_weights__()

    def __init_weights__(self):
        nn.init.xavier_normal_(self.W_q)
        nn.init.xavier_normal_(self.W_k)
        nn.init.xavier_normal_(self.W_v)
        nn.init.xavier_normal_(self.W_o)

        nn.init.xavier_normal_(self.ffn.dense1.weight)
        nn.init.xavier_normal_(self.ffn.dense2.weight)

    def multi_head_attention(self, Q, K, V):
        # 64 * 1 * self.embedding_size，进去是什么形状，出来就是什么形状
        batch_size, q_len, _ = Q.size()
        batch_size, k_len, _ = K.size()
        batch_size, v_len, _ = V.size()
        Q_ = transpose_qkv(Q.matmul(self.W_q), self.head_num)
        K_ = transpose_qkv(K.matmul(self.W_k), self.head_num)
        V_ = transpose_qkv(V.matmul(self.W_v), self.head_num)

        attention_score = self.dot_product_attention(Q_, K_, V_)
        attention_score = transpose_output(attention_score, self.head_num)

        output = self.dropout(attention_score.matmul(self.W_o))
        return output

    def forward(self, Q, K, V):
        """
        only for self-attention, the input dimensions of Q, K, V are the same

        :param Q: (batch_size, max_q_words, input_size)
        :param K: (batch_size, max_k_words, input_size)
        :param V: (batch_size, max_v_words, input_size)
        :return: output: (batch_size, max_q_words, input_size)  same size as Q
        """
        attention_score = self.multi_head_attention(Q, K, V)
        # without norm
        X = Q + attention_score
        output = self.ffn(X) + X
        return output


class MFAN(AbstractModel):
    """
    MFAN: Multi-modal Feature-enhanced TransformerBlock Networks for Rumor Detection
    """
    def __init__(self,
                 word_vectors: torch.Tensor,
                 node_num: int,
                 node_embedding: torch.Tensor,
                 adj_matrix: torch.Tensor,
                 dropout_rate=0.6):
        """

        Args:
            word_vectors (Tensor): pretrained weights for word embedding
            node_num (int): number of nodes in graph
            node_embedding (Tensor): pretrained weights for node embedding
            adj_matrix (Tensor): adjacent matrix of graph
            dropout_rate (float): drop out rate. Default=0.6
        """
        super(MFAN, self).__init__()

        # text embedding
        self.word_embedding = nn.Embedding.from_pretrained(word_vectors, freeze=False, padding_idx=0)
        self.embedding_size = word_vectors.shape[-1]

        # text cnn
        kernel_sizes = [3, 4, 5]
        self.text_cnn_layer = TextCNNLayer(self.embedding_size, 100,
                                           kernel_sizes, nn.ReLU())

        # graph
        self.node_num = node_num
        self.cos_matrix = calculate_cos_matrix(node_embedding, node_embedding)
        self.signed_gat = SignedGAT(node_vectors=node_embedding,
                                    cos_sim_matrix=self.cos_matrix,
                                    num_features=self.embedding_size,
                                    node_num=self.node_num,
                                    head_num=1,
                                    adj_matrix=adj_matrix,
                                    dropout=0)

        # image
        self.resnet = models.resnet50(weights=models.ResNet50_Weights.DEFAULT)
        self.resnet.fc = nn.Linear(2048, self.embedding_size)
        nn.init.eye_(self.resnet.fc.weight)

        # feature fusion
        self.align_graph = nn.Linear(self.embedding_size, self.embedding_size)
        self.align_text = nn.Linear(self.embedding_size, self.embedding_size)
        self.transformer_block = TransformerBlock(
            input_size=self.embedding_size, head_num=8, dropout=0)

        # classification
        self.dropout = nn.Dropout(dropout_rate)
        self.relu = nn.ReLU()
        self.classifier = nn.Sequential(nn.Linear(1800, 900),
                                        self.dropout, self.relu,
                                        nn.Linear(900, 600), self.relu,
                                        nn.Linear(600, self.embedding_size),
                                        self.relu, self.dropout,
                                        nn.Linear(self.embedding_size, 2))
        self.__init_weights__()

        self.loss_funcs = [nn.CrossEntropyLoss(), nn.MSELoss()]

    def __init_weights__(self):
        for module in self.classifier:
            if type(module) is torch.nn.Linear:
                nn.init.xavier_normal_(module.weight)

    def forward(self, post_id: torch.Tensor, text: torch.Tensor,
                image: torch.Tensor):
        """

        Args:
            post_id (Tensor): id of post, shape=(batch_size,)
            text (Tensor): token ids, shape=(batch_size, max_len)
            image (Tensor): shape=(batch_size, 3, width, height)

        Returns:
            tuple:
                - class_output (Tensor): prediction of being fake news, shape=(batch_size, 2)
                - dist (List[Tensor]): aligned text and aligned graph, shape=(batch_size, embedding_size)
        """
        image_feature = self.resnet(image).unsqueeze(1)

        graph_feature = self.signed_gat.forward(post_id).unsqueeze(1)

        # text
        text = self.word_embedding(text)
        text = self.text_cnn_layer(text)
        text_feature = text.unsqueeze(1)

        # text image graph各自self-attention
        self_att_t = self.transformer_block(text_feature, text_feature,
                                            text_feature)
        self_att_g = self.transformer_block(graph_feature, graph_feature,
                                            graph_feature)
        self_att_i = self.transformer_block(image_feature, image_feature,
                                            image_feature)

        # 将Q换成image，text image co-attention，得到增强后的text
        enhanced_text = self.transformer_block(self_att_i, self_att_t,
                                               self_att_t)

        # 此后使用的text，均为enhanced，而非最开始self-attention的版本
        self_att_t = enhanced_text

        # enhanced text与enhanced graph对齐
        aligned_text = self.align_text(enhanced_text.squeeze(1))
        aligned_graph = self.align_graph(self_att_g.squeeze(1))
        dist = [aligned_text, aligned_graph]

        # co attention
        co_att_tg = self.transformer_block(self_att_t, self_att_g,
                                           self_att_g).squeeze(1)
        co_att_gt = self.transformer_block(self_att_g, self_att_t,
                                           self_att_t).squeeze(1)
        co_att_ti = self.transformer_block(self_att_t, self_att_i,
                                           self_att_i).squeeze(1)
        co_att_it = self.transformer_block(self_att_i, self_att_t,
                                           self_att_t).squeeze(1)
        co_att_gi = self.transformer_block(self_att_g, self_att_i,
                                           self_att_i).squeeze(1)
        co_att_ig = self.transformer_block(self_att_i, self_att_g,
                                           self_att_g).squeeze(1)

        #  最终分类
        att_feature = torch.cat(
            (co_att_tg, co_att_gt, co_att_ti, co_att_it, co_att_gi, co_att_ig),
            dim=1)
        class_output = self.classifier(att_feature)

        return class_output, dist

    def calculate_loss(self, data) -> Dict[str, Tensor]:
        post_id = data['post_id']
        text = data['text']
        image = data['image']
        label = data['label']
        class_output, dist = self.forward(post_id, text, image)
        class_loss = self.loss_funcs[0](class_output, label)
        dis_loss = self.loss_funcs[1](dist[0], dist[1])

        loss = class_loss + dis_loss

        return {'total_loss': loss, 'class_loss': class_loss, 'dis_loss': dis_loss}

    def predict(self, data_without_label):
        post_id = data_without_label['post_id']
        text = data_without_label['text']
        image = data_without_label['image']
        class_output, _ = self.forward(post_id, text, image)
        return class_output



import datetime
import random
import warnings
from collections import defaultdict
from typing import List, Dict

import numpy as np
import torch
from sklearn.metrics.pairwise import cosine_similarity


def calculate_cos_matrix(matrix1: torch.Tensor, matrix2: torch.Tensor):
    # pairwise 计算matrix1中每行向量与matrix2中所有行向量的余弦相似度
    return torch.from_numpy(cosine_similarity(matrix1.numpy(),
                                              matrix2.numpy()))


def dict2str(result_dict: Dict[str, float]) -> str:
    r"""convert result dict to str

    Args:
        result_dict (dict): result dict

    Returns:
        str: result str
    """

    return "    ".join([
        str(metric) + "=" + f"{value:.6f}"
        for metric, value in result_dict.items()
    ])


def now2str() -> str:
    r"""convert current time to str

    Returns:
        str: current time, %Y-%m-%d-%H_%M_%S
    """
    cur = datetime.datetime.now()
    cur = cur.strftime("%Y-%m-%d-%H_%M_%S")

    return cur


def seconds2str(seconds: float) -> str:
    r"""Convert seconds to time format"""
    if seconds < 60:
        return f'{seconds:.6f}s'

    m, s = divmod(int(seconds), 60)
    h, m = divmod(m, 60)
    if m == 0:
        return f'{s}s'
    elif h == 0:
        return f'{m}m{s}s'
    return f'{h}h{m}m{s}s'


def check_loss_type(result):
    result_is_dict = False

    if type(result) is dict:
        result_is_dict = True
        if 'total_loss' in result.keys():
            loss = result['total_loss']
        else:
            # todo 是否允许没有total_loss，采用所有loss的和作为total_loss
            warnings.warn(
                f"no total_loss in result: {result}, use sum of all losses as total_loss"
            )
            loss = torch.sum(torch.stack(list(result.values())))
    elif type(result) is torch.Tensor:
        loss = result
    else:
        raise TypeError(f"result type error: {result}")

    return loss, result_is_dict


def lsh_data_selection(domain_embeddings: torch.Tensor, labelling_budget=100, hash_dimension=10) -> List[int]:
    """
    local sensitive hash selection for training dataset
    Args:
        domain_embeddings (Tensor): 2-D domain embedding tensor of samples
        labelling_budget (int): number of selection budget, must be smaller than number of samples. Default=100
        hash_dimension (int): dimension of random hash vector. Default=10

    Returns:
        List[int], a list of selected samples index
    """
    if len(domain_embeddings.shape) != 2:
        raise TypeError("domain embedding must be 2-D tensor!")

    if labelling_budget > domain_embeddings.shape[0]:
        raise RuntimeError(
            f"labelling budget({labelling_budget}) is greater than data pool size({domain_embeddings.shape[0]})")

    embedding_size = domain_embeddings.shape[1]
    final_selected_ids = []
    is_final_selected = defaultdict(lambda: False)

    random_distribution = [3 ** 0.5, 0.0, 0.0, 0.0, 0.0, -(3 ** 0.5)]

    while len(final_selected_ids) < labelling_budget:
        # generate random vectors
        random_vectors = []
        for hash_run in range(hash_dimension):
            vec = random.choices(random_distribution, k=embedding_size)
            random_vectors.append(torch.tensor(vec))

        # create hash table
        code_dict = defaultdict(lambda: [])  # {str(h-dim hash value): [domain_id]}
        for i, item in enumerate(domain_embeddings):
            code = ''
            # skip if the item is already selected
            if is_final_selected[i]:
                continue

            # concat result of hash functions to generate hash vectors
            for code_vec in random_vectors:
                code = code + str(int(torch.dot(item, code_vec) > 0))
            code_dict[code].append(i)

        selected_ids = []
        is_selected = defaultdict(lambda: False)

        # pick one item from each item bin
        for item in code_dict:
            selected_item = random.choice(code_dict[item])
            selected_ids.append(selected_item)  # 添加domain id，即domain embedding中的行号
            is_selected[selected_item] = True

        # remove a set of instances randomly to meet the labelling budget
        if len(final_selected_ids + selected_ids) > labelling_budget:
            random_pick_size = labelling_budget - len(final_selected_ids)
            mod_selected_ids = []
            for z in range(random_pick_size):
                select_item = random.choice(selected_ids)
                mod_selected_ids.append(select_item)
                selected_ids.remove(select_item)
            final_selected_ids = final_selected_ids + mod_selected_ids
            return final_selected_ids

        for item in selected_ids:
            is_final_selected[item] = True
        final_selected_ids = final_selected_ids + selected_ids
    return final_selected_ids


def read_stop_words(path: str) -> List[str]:
    with open(path, 'r', encoding='utf-8') as f:
        stop_words = [str(line).strip() for line in f.readlines()]
    return stop_words


class EarlyStopping:
    """
    Early stopping to stop the training when the score does not improve after
    certain epochs.
    """

    def __init__(self, patience=10, delta=0.000001, mode='max'):
        """
        Args:
            patience (int): number of epochs to wait for improvement, default=10
            delta (float): minimum change in the monitored quantity to qualify as an improvement, default=0.000001
            mode (str): minimize or maximize score, one of {min, max}, default=max
        """

        assert mode in ['min', 'max'], "mode must be either 'min' or 'max'"

        self.patience = patience
        self.delta = delta
        self.counter = 0
        self.best_score = None
        self.early_stop = False
        self.mode = mode

    def __call__(self, current_score: float) -> bool:
        """
        Args:
            current_score (float): current score to check if it is the best score

        Returns:
            bool: whether the current score is the best score
        """
        improvement = False
        if self.best_score is None:
            improvement = True
        elif self.mode == 'min':
            if current_score < self.best_score - self.delta:
                improvement = True
        elif self.mode == 'max':
            if current_score > self.best_score + self.delta:
                improvement = True

        if improvement:
            self.best_score = current_score
            self.counter = 0
        else:
            self.counter += 1
            if self.counter >= self.patience:
                self.early_stop = True

        return improvement


class DropEdge:
    def __init__(self, td_drop_rate, bu_drop_rate):
        """
            Drop edge operation from BiGCN (Rumor Detection on Social Media with Bi-Directional Graph Convolutional Networks)
            1) Generate TD and BU edge indices
            2) Drop out edges
            Code from https://github.com/TianBian95/BiGCN/blob/master/Process/dataset.py
        """
        self.td_drop_rate = td_drop_rate
        self.bu_drop_rate = bu_drop_rate

    def __call__(self, data):
        edge_index = data.edge_index

        if self.td_drop_rate > 0:
            row = list(edge_index[0])
            col = list(edge_index[1])
            length = len(row)
            poslist = random.sample(range(length),
                                    int(length * (1 - self.td_drop_rate)))
            poslist = sorted(poslist)
            row = list(np.array(row)[poslist])
            col = list(np.array(col)[poslist])
            new_edge_index = [row, col]
        else:
            new_edge_index = edge_index

        bu_row = list(edge_index[1])
        bu_col = list(edge_index[0])
        if self.bu_drop_rate > 0:
            length = len(bu_row)
            poslist = random.sample(range(length),
                                    int(length * (1 - self.bu_drop_rate)))
            poslist = sorted(poslist)
            row = list(np.array(bu_row)[poslist])
            col = list(np.array(bu_col)[poslist])
            bu_new_edge_index = [row, col]
        else:
            bu_new_edge_index = [bu_row, bu_col]

        data.edge_index = torch.LongTensor(new_edge_index)
        data.BU_edge_index = torch.LongTensor(bu_new_edge_index)
        data.root = torch.FloatTensor(data.x[0])
        data.root_index = torch.LongTensor([0])

        return data
